import yaml

from datetime import datetime



import os
import torch
import torch.nn as nn
import torch.optim as optim
import matplotlib.pyplot as plt

from torch.utils.data import DataLoader

from best.dbs.artifact_removal.model import *
from best.dbs.artifact_removal.dataset import *
from best.dbs import *
from best.dbs.artifact_removal.model import dbs_artifact_removal_network_v2

from best._config import ObjDictToDict

torch.autograd.set_detect_anomaly(True)

class TrainerUpgrade:
    def __init__(self, config):
        self.cfg = config
        self.model_name = self.cfg.NAME
        self.fs = self.cfg.MODEL.ARCHITECTURE.FS
        exec(
        f"self.DatasetTrain={self.cfg.TRAIN.DATASET.split('.')[-1]}({self.cfg.TRAIN.SIGNAL_LENGTH}, {self.cfg.TRAIN.EEG_GENERATORS_DCGAN}, {self.fs}, {self.cfg.TRAIN.STIM_ARTIFACT})"
        )
        self.DatasetTrain._len = self.cfg.TRAIN.BATCH_SIZE * self.cfg.TRAIN.N_ITERATIONS

        self.path_report = os.path.join(self.cfg.TRAIN.PATH_REPORT, self.cfg.NAME + "_" + datetime.now().strftime("%Y_%m_%d_%H_%M_%S"))
        self.path_report_models = os.path.join(self.path_report, 'Models')
        self.path_report_images = os.path.join(self.path_report, 'Images')
        self.path_report_config = os.path.join(self.path_report, 'config.yaml')
        self.path_report_losses = os.path.join(self.path_report, 'losses.csv')
        self.path_report_losses_validation = os.path.join(self.path_report, 'losses_val.csv')

        os.mkdir(self.path_report)
        os.mkdir(self.path_report_models)
        os.mkdir(self.path_report_images)
        with open(self.path_report_config, 'w') as file:
            yaml.dump(ObjDictToDict(self.cfg), file)

        with open(self.path_report_losses, 'w') as f:
            f.write('Epoch, lr, loss_rec_signal, loss_rec_spectrogram, loss_detection\n')

        self.save_model_freq = self.cfg.TRAIN.SAVE_MODEL_EPOCH
        self.save_report_freq = self.cfg.TRAIN.SAVE_REPORT_EPOCH

        self.GPUs = self.cfg.TRAIN.GPU
        self.num_gpus = self.GPUs.__len__()
        self.minibatchsize = int(self.cfg.TRAIN.BATCH_SIZE)

        self.lr = self.cfg.TRAIN.BASE_LR

        self.epochs = self.cfg.TRAIN.EPOCHS


        self.current_epoch = 0
        self.current_iteration = 0
        self.overall_iteration = 0
        self.epoch_list = []
        self.lossKL_list = []
        self.lossMSE_list = []
        self.loss_list = []
        self.device = self.GPUs[0]

        self.model = dbs_artifact_removal_network_v2()
        self.model.cuda(self.GPUs[0])

        # exec(
        # f"self.model={self.cfg.MODEL.MODEL.split('.')[-1]}(n_filters={self.cfg.MODEL.ARCHITECTURE.N_FILTERS})"
        # )


        self.optimizer = optim.Adam(self.model.parameters(), lr=self.cfg.TRAIN.BASE_LR, betas=(self.cfg.TRAIN.BETA_1, 0.999))
        self.lr_scheduler = torch.optim.lr_scheduler.MultiStepLR(self.optimizer, milestones=self.cfg.TRAIN.DECAY_ITERATIONS, gamma=self.cfg.TRAIN.DECAY_FACTOR)

        self.criterionMSE = nn.MSELoss()
        self.criterionBCE = nn.BCELoss()

        def weights_init(m):
            classname = m.__class__.__name__
            if classname.find('Conv') != -1:
                nn.init.normal_(m.weight.data, 0.0, 0.02)
            elif classname.find('BatchNorm') != -1:
                nn.init.normal_(m.weight.data, 1.0, 0.02)
                nn.init.constant_(m.bias.data, 0)

        # self.model.apply(weights_init)
        self.DLoaderTrain = DataLoader(self.DatasetTrain, batch_size=self.minibatchsize, shuffle=True, num_workers=self.cfg.TRAIN.CPU_COUNT_LOADERS, drop_last=False)
        # self.model.cuda(self.device)


    def do_epoch(self):
        self.current_iteration = 0
        print('----------------------------------')
        print('VAE Epoch ' + str(self.current_epoch))


        self.model.train()
        for k, (x_orig, x_art, y_art) in enumerate(self.DLoaderTrain):
            # break
            self.model.zero_grad()
            x_orig = x_orig.float().detach().to(self.device)
            x_art = x_art.float().detach().to(self.device)
            y_art = y_art.float().detach().to(self.device)

            x_art_diff_1 = x_art.diff(1, -1, x_art[:, :, 0].view(x_art.shape[0], x_art.shape[1], 1))
            x_art_diff_2 = x_art_diff_1.diff(1, -1, x_art_diff_1[:, :, 0].view(x_art_diff_1.shape[0], x_art_diff_1.shape[1], 1))

            x_art_inp = torch.cat(
                (
                    x_art,
                    x_art_diff_1,
                    x_art_diff_2
                ), 1
            )

            x_rec, att = self.model(x_art_inp)
            #x_orig_fft = torch.stft(x_orig.squeeze(1), n_fft=2*500, hop_length=100, return_complex=True, ).abs()
            #x_rec_fft = torch.stft(x_rec.squeeze(1), n_fft=2*500, hop_length=100, return_complex=True, ).abs()
            # break
            x_orig_fft = torch.fft.fft(x_orig.squeeze(1)).abs()
            x_rec_fft = torch.fft.fft(x_rec.squeeze(1)).abs()

            x_orig_fft = torch.log10(x_orig_fft+0.0001)
            x_rec_fft = torch.log10(x_rec_fft+0.0001)

            x_rec_fft[torch.isnan(x_orig_fft) | torch.isinf(x_orig_fft)] = 0
            x_orig_fft[torch.isnan(x_orig_fft) | torch.isinf(x_orig_fft)] = 0

            x_orig_fft[torch.isnan(x_rec_fft) | torch.isinf(x_rec_fft)] = 0
            x_rec_fft[torch.isnan(x_rec_fft) | torch.isinf(x_rec_fft)] = 0


            #att[att < 0] = att[att < 0]*0
            #att[att > 1] = att[att > 1]/att[att > 1]


            loss_signal = self.criterionMSE(x_rec[~torch.isnan(x_rec)], x_orig[~torch.isnan(x_rec)]) * self.cfg.TRAIN.WEIGHT_SIGNAL_RECONSTRUCTION
            loss_fft = self.criterionMSE(x_rec_fft[~torch.isnan(x_rec_fft)], x_orig_fft[~torch.isnan(x_rec_fft)])* self.cfg.TRAIN.WEIGHT_STFT_RECONSTRUCTION
            loss_att = self.criterionBCE(att[~torch.isnan(att)], y_art[~torch.isnan(att)]) * self.cfg.TRAIN.WEIGHT_DETECTION


            loss = loss_signal + loss_fft #+ loss_att
            loss.backward()

            torch.nn.utils.clip_grad_value_(self.model.parameters(), 0.1)
            self.optimizer.step()
            self.lr_scheduler.step()


            losses = {
                'loss_rec_signal': loss_signal.item(),
                'loss_rec_spectrogram': loss_fft.item(),
                'loss_detection': loss_att.item(),
            }


            if self.overall_iteration % self.cfg.TRAIN.SAVE_REPORT_ITERATION == 0 or k==0:
                self.print_losses_to_file(losses)
                self.plot_to_file(x_orig, x_art, x_rec, y_art, att)

            if self.overall_iteration % self.cfg.TRAIN.SAVE_MODEL_ITERATION == 0 or k==0:
                self.save_model()

            self.current_iteration += 1
            self.overall_iteration += 1
        self.current_epoch += 1


    def print_losses_to_file(self, losses):
        path = self.path_report_losses
        lr = self.optimizer.param_groups[0]['lr']
        printStr = f'{self.current_iteration}, {lr}, ' + ', '.join([f'{i:.8f}' for k, i in losses.items()]) + '\n'
        print(self.overall_iteration, self.current_epoch, self.current_iteration, losses)
        with open(path, 'a') as f:
            f.write(printStr)

    # (x_orig, x_rec, y_art, att)
    def plot_to_file(self, x_orig, x_art, x_rec, y_art, yy_art):
        x_orig = x_orig[-1].detach().cpu().squeeze().numpy()
        x_rec = x_rec[-1].detach().cpu().squeeze().numpy()
        x_art = x_art[-1].detach().cpu().squeeze().numpy()
        y_art = y_art[-1].detach().cpu().squeeze().numpy()
        yy_art = yy_art[-1].detach().cpu().squeeze().numpy()
        img_path = os.path.join(self.path_report_images, f"{self.cfg.NAME}_epoch_{self.current_epoch:05d}_iteration_{self.current_iteration:05d}_step_{self.overall_iteration:05d}")

        plt.figure(figsize=(12, 4))
        ax0 = plt.subplot(3, 1, 1)
        plt.plot(x_orig)
        plt.plot(x_rec)
        plt.subplot(3, 1, 2, sharex=ax0)
        plt.plot(x_art)
        plt.subplot(3, 1, 3, sharex=ax0)
        plt.plot(y_art)
        plt.plot(yy_art)
        plt.savefig(img_path + '.png')
        # plt.savefig(img_path + '.pdf')
        plt.close()
        plt.show()

    def save_model(self):
        PATH = os.path.join(self.path_report_models, f"{self.cfg.NAME}_epoch_{self.current_epoch:05d}_iteration_{self.current_iteration:05d}_step_{self.overall_iteration:05d}_ArtifactEraser.pt")
        # torch.save(self.model.state_dict(), PATH)

    def train(self):
        self.do_epoch()


class Trainer:
    def __init__(self, config):
        self.cfg = config
        self.model_name = self.cfg.NAME
        self.fs = self.cfg.MODEL.ARCHITECTURE.FS
        exec(
        f"self.DatasetTrain={self.cfg.TRAIN.DATASET.split('.')[-1]}({self.cfg.TRAIN.SIGNAL_LENGTH}, {self.cfg.TRAIN.EEG_GENERATORS_DCGAN}, {self.fs}, {self.cfg.TRAIN.STIM_ARTIFACT})"
        )
        self.DatasetTrain._len = self.cfg.TRAIN.BATCH_SIZE * self.cfg.TRAIN.N_ITERATIONS

        self.path_report = os.path.join(self.cfg.TRAIN.PATH_REPORT, self.cfg.NAME + "_" + datetime.now().strftime("%Y_%m_%d_%H_%M_%S"))
        self.path_report_models = os.path.join(self.path_report, 'Models')
        self.path_report_images = os.path.join(self.path_report, 'Images')
        self.path_report_config = os.path.join(self.path_report, 'config.yaml')
        self.path_report_losses = os.path.join(self.path_report, 'losses.csv')
        self.path_report_losses_validation = os.path.join(self.path_report, 'losses_val.csv')

        os.mkdir(self.path_report)
        os.mkdir(self.path_report_models)
        os.mkdir(self.path_report_images)
        with open(self.path_report_config, 'w') as file:
            yaml.dump(ObjDictToDict(self.cfg), file)

        with open(self.path_report_losses, 'w') as f:
            f.write('Epoch, lr, loss_rec_signal, loss_rec_spectrogram, loss_detection\n')

        self.save_model_freq = self.cfg.TRAIN.SAVE_MODEL_EPOCH
        self.save_report_freq = self.cfg.TRAIN.SAVE_REPORT_EPOCH

        self.GPUs = self.cfg.TRAIN.GPU
        self.num_gpus = self.GPUs.__len__()
        self.minibatchsize = int(self.cfg.TRAIN.BATCH_SIZE)

        self.lr = self.cfg.TRAIN.BASE_LR

        self.epochs = self.cfg.TRAIN.EPOCHS


        self.current_epoch = 0
        self.current_iteration = 0
        self.overall_iteration = 0
        self.epoch_list = []
        self.lossKL_list = []
        self.lossMSE_list = []
        self.loss_list = []
        self.device = self.GPUs[0]

        #self.model = GRU_Denoiser(n_filters=self.cfg.MODEL.ARCHITECTURE.N_FILTERS)

        exec(
        f"self.model={self.cfg.MODEL.MODEL.split('.')[-1]}(n_filters={self.cfg.MODEL.ARCHITECTURE.N_FILTERS})"
        )


        self.optimizer = optim.Adam(self.model.parameters(), lr=self.cfg.TRAIN.BASE_LR, betas=(self.cfg.TRAIN.BETA_1, 0.999))
        self.lr_scheduler = torch.optim.lr_scheduler.MultiStepLR(self.optimizer, milestones=self.cfg.TRAIN.DECAY_ITERATIONS, gamma=self.cfg.TRAIN.DECAY_FACTOR)

        self.criterionMSE = nn.MSELoss()
        self.criterionBCE = nn.BCELoss()

        def weights_init(m):
            classname = m.__class__.__name__
            if classname.find('Conv') != -1:
                nn.init.normal_(m.weight.data, 0.0, 0.02)
            elif classname.find('BatchNorm') != -1:
                nn.init.normal_(m.weight.data, 1.0, 0.02)
                nn.init.constant_(m.bias.data, 0)

        self.model.apply(weights_init)
        self.DLoaderTrain = DataLoader(self.DatasetTrain, batch_size=self.minibatchsize, shuffle=True, num_workers=self.cfg.TRAIN.CPU_COUNT_LOADERS, drop_last=False)
        self.model.cuda(self.device)


    def do_epoch(self):
        self.current_iteration = 0
        print('----------------------------------')
        print('VAE Epoch ' + str(self.current_epoch))


        self.model.train()
        #print('DOPICE')
        for k, (x_orig, x_art, y_art) in enumerate(self.DLoaderTrain):
            #break
            self.model.zero_grad()
            x_orig = x_orig.float().detach().to(self.device)
            x_art = x_art.float().detach().to(self.device)
            y_art = y_art.float().detach().to(self.device)

            x_rec, att = self.model(x_art)
            #x_orig_fft = torch.stft(x_orig.squeeze(1), n_fft=2*500, hop_length=100, return_complex=True, ).abs()
            #x_rec_fft = torch.stft(x_rec.squeeze(1), n_fft=2*500, hop_length=100, return_complex=True, ).abs()

            x_orig_fft = torch.fft.fft(x_orig.squeeze(1)).abs()
            x_rec_fft = torch.fft.fft(x_rec.squeeze(1)).abs()

            x_orig_fft = torch.log10(x_orig_fft+0.0001)
            x_rec_fft = torch.log10(x_rec_fft+0.0001)

            x_rec_fft[torch.isnan(x_orig_fft) | torch.isinf(x_orig_fft)] = 0
            x_orig_fft[torch.isnan(x_orig_fft) | torch.isinf(x_orig_fft)] = 0

            x_orig_fft[torch.isnan(x_rec_fft) | torch.isinf(x_rec_fft)] = 0
            x_rec_fft[torch.isnan(x_rec_fft) | torch.isinf(x_rec_fft)] = 0


            #att[att < 0] = att[att < 0]*0
            #att[att > 1] = att[att > 1]/att[att > 1]


            loss_signal = self.criterionMSE(x_rec[~torch.isnan(x_rec)], x_orig[~torch.isnan(x_rec)]) * self.cfg.TRAIN.WEIGHT_SIGNAL_RECONSTRUCTION
            loss_fft = self.criterionMSE(x_rec_fft[~torch.isnan(x_rec_fft)], x_orig_fft[~torch.isnan(x_rec_fft)])* self.cfg.TRAIN.WEIGHT_STFT_RECONSTRUCTION
            loss_att = self.criterionBCE(att[~torch.isnan(att)], y_art[~torch.isnan(att)]) * self.cfg.TRAIN.WEIGHT_DETECTION
            # loss_att = self.criterionMSE(x_art[~torch.isnan(x_rec)], att[~torch.isnan(x_rec)]) * self.cfg.TRAIN.WEIGHT_DETECTION


            loss = loss_signal + loss_att + loss_fft
            loss.backward()

            torch.nn.utils.clip_grad_value_(self.model.parameters(), 0.1)
            self.optimizer.step()
            self.lr_scheduler.step()


            losses = {
                'loss_rec_signal': loss_signal.item(),
                'loss_rec_spectrogram': loss_fft.item(),
                'loss_detection': loss_att.item(),
            }


            if self.overall_iteration % self.cfg.TRAIN.SAVE_REPORT_ITERATION == 0 or k==0:
                self.print_losses_to_file(losses)
                self.plot_to_file(x_orig, x_art, x_rec, y_art, att)

            if self.overall_iteration % self.cfg.TRAIN.SAVE_MODEL_ITERATION == 0 or k==0:
                self.save_model()

            self.current_iteration += 1
            self.overall_iteration += 1

        self.current_epoch += 1


    def print_losses_to_file(self, losses):
        path = self.path_report_losses
        lr = self.optimizer.param_groups[0]['lr']
        printStr = f'{self.current_iteration}, {lr}, ' + ', '.join([f'{i:.8f}' for k, i in losses.items()]) + '\n'
        print(self.overall_iteration, self.current_epoch, self.current_iteration, losses)
        with open(path, 'a') as f:
            f.write(printStr)

    # (x_orig, x_rec, y_art, att)
    def plot_to_file(self, x_orig, x_art, x_rec, y_art, yy_art):
        x_orig = x_orig[-1].detach().cpu().squeeze().numpy()
        x_rec = x_rec[-1].detach().cpu().squeeze().numpy()
        x_art = x_art[-1].detach().cpu().squeeze().numpy()
        y_art = y_art[-1].detach().cpu().squeeze().numpy()
        yy_art = yy_art[-1].detach().cpu().squeeze().numpy()
        img_path = os.path.join(self.path_report_images, f"{self.cfg.NAME}_epoch_{self.current_epoch:05d}_iteration_{self.current_iteration:05d}_step_{self.overall_iteration:05d}")

        plt.figure(figsize=(12, 4))
        ax0 = plt.subplot(3, 1, 1)
        plt.plot(x_orig)
        plt.plot(x_rec)
        plt.subplot(3, 1, 2, sharex=ax0)
        plt.plot(x_art)
        plt.subplot(3, 1, 3, sharex=ax0)
        plt.plot(y_art)
        plt.plot(yy_art)
        plt.savefig(img_path + '.png')
        plt.savefig(img_path + '.pdf')
        plt.close()
        #plt.show()

    def save_model(self):
        PATH = os.path.join(self.path_report_models, f"{self.cfg.NAME}_epoch_{self.current_epoch:05d}_iteration_{self.current_iteration:05d}_step_{self.overall_iteration:05d}_ArtifactEraser.pt")
        torch.save(self.model.state_dict(), PATH)

    def train(self):
        self.do_epoch()


class TrainerLight:
    def __init__(self, config):
        self.cfg = config
        self.model_name = self.cfg.NAME
        exec(
        f"self.DatasetTrain={self.cfg.TRAIN.DATASET.split('.')[-1]}({self.cfg.TRAIN.SIGNAL_LENGTH}, {self.cfg.TRAIN.EEG_GENERATORS_DCGAN}, {self.cfg.TRAIN.STIM_ARTIFACT})"
        )
        self.DatasetTrain._len = self.cfg.TRAIN.BATCH_SIZE * self.cfg.TRAIN.N_ITERATIONS


        self.path_report = os.path.join(self.cfg.TRAIN.PATH_REPORT, self.cfg.NAME + "_" + datetime.now().strftime("%Y_%m_%d_%H_%M_%S"))
        self.path_report_models = os.path.join(self.path_report, 'Models')
        self.path_report_images = os.path.join(self.path_report, 'Images')
        self.path_report_config = os.path.join(self.path_report, 'config.yaml')
        self.path_report_losses = os.path.join(self.path_report, 'losses.csv')
        self.path_report_losses_validation = os.path.join(self.path_report, 'losses_val.csv')


        # os.mkdir(self.path_report)
        # os.mkdir(self.path_report_models)
        # os.mkdir(self.path_report_images)
        # with open(self.path_report_config, 'w') as file:
            # yaml.dump(ObjDictToDict(self.cfg), file)


        # with open(self.path_report_losses, 'w') as f:
        #     f.write('Epoch, lr, loss_rec_signal, loss_rec_spectrogram, loss_detection\n')

        self.save_model_freq = self.cfg.TRAIN.SAVE_MODEL_EPOCH
        self.save_report_freq = self.cfg.TRAIN.SAVE_REPORT_EPOCH

        self.GPUs = self.cfg.TRAIN.GPU
        self.num_gpus = self.GPUs.__len__()
        self.minibatchsize = int(self.cfg.TRAIN.BATCH_SIZE)

        self.lr = self.cfg.TRAIN.BASE_LR

        self.epochs = self.cfg.TRAIN.EPOCHS


        self.current_epoch = 0
        self.current_iteration = 0
        self.overall_iteration = 0
        self.epoch_list = []
        self.lossKL_list = []
        self.lossMSE_list = []
        self.loss_list = []
        self.device = self.GPUs[0]

        self.model = dbs_artifact_removal_network(n_filters=self.cfg.MODEL.ARCHITECTURE.N_FILTERS)

        exec(
        f"self.model={self.cfg.MODEL.MODEL.split('.')[-1]}(n_filters={self.cfg.MODEL.ARCHITECTURE.N_FILTERS})"
        )


        self.optimizer = optim.Adam(self.model.parameters(), lr=self.cfg.TRAIN.BASE_LR, betas=(self.cfg.TRAIN.BETA_1, 0.999))
        self.lr_scheduler = torch.optim.lr_scheduler.MultiStepLR(self.optimizer, milestones=self.cfg.TRAIN.DECAY_ITERATIONS, gamma=self.cfg.TRAIN.DECAY_FACTOR)

        self.criterionMSE = nn.MSELoss()
        self.criterionBCE = nn.BCELoss()

        def weights_init(m):
            classname = m.__class__.__name__
            if classname.find('Conv') != -1:
                nn.init.normal_(m.weight.data, 0.0, 0.02)
            elif classname.find('BatchNorm') != -1:
                nn.init.normal_(m.weight.data, 1.0, 0.02)
                nn.init.constant_(m.bias.data, 0)

        self.model.apply(weights_init)
        self.DLoaderTrain = DataLoader(self.DatasetTrain, batch_size=self.minibatchsize, shuffle=True, num_workers=self.cfg.TRAIN.CPU_COUNT_LOADERS, drop_last=False)
        self.model.cuda(self.device)


    def do_epoch(self):
        self.current_iteration = 0
        print('----------------------------------')
        print('VAE Epoch ' + str(self.current_epoch))


        self.model.train()
        #print('DOPICE')
        for k, (x_orig, x_art, y_art) in enumerate(self.DLoaderTrain):
            #break
            self.model.zero_grad()
            x_orig = x_orig.float().detach().to(self.device)
            x_art = x_art.float().detach().to(self.device)
            y_art = y_art.float().detach().to(self.device)

            x_rec = self.model(x_art)
            #x_orig_fft = torch.stft(x_orig.squeeze(1), n_fft=2*500, hop_length=100, return_complex=True, ).abs()
            #x_rec_fft = torch.stft(x_rec.squeeze(1), n_fft=2*500, hop_length=100, return_complex=True, ).abs()

            x_orig_fft = torch.fft.fft(x_orig.squeeze(1)).abs()
            x_rec_fft = torch.fft.fft(x_rec.squeeze(1)).abs()

            x_orig_fft = torch.log10(x_orig_fft+0.0001)
            x_rec_fft = torch.log10(x_rec_fft+0.0001)

            x_rec_fft[torch.isnan(x_orig_fft) | torch.isinf(x_orig_fft)] = 0
            x_orig_fft[torch.isnan(x_orig_fft) | torch.isinf(x_orig_fft)] = 0

            x_orig_fft[torch.isnan(x_rec_fft) | torch.isinf(x_rec_fft)] = 0
            x_rec_fft[torch.isnan(x_rec_fft) | torch.isinf(x_rec_fft)] = 0


            #att[att < 0] = att[att < 0]*0
            #att[att > 1] = att[att > 1]/att[att > 1]


            loss_signal = self.criterionMSE(x_rec[~torch.isnan(x_rec)], x_orig[~torch.isnan(x_rec)]) * self.cfg.TRAIN.WEIGHT_SIGNAL_RECONSTRUCTION
            loss_fft = self.criterionMSE(x_rec_fft[~torch.isnan(x_rec_fft)], x_orig_fft[~torch.isnan(x_rec_fft)])* self.cfg.TRAIN.WEIGHT_STFT_RECONSTRUCTION


            loss = loss_signal + loss_fft
            loss.backward()

            torch.nn.utils.clip_grad_value_(self.model.parameters(), 0.1)
            self.optimizer.step()
            self.lr_scheduler.step()


            losses = {
                'loss_rec_signal': loss_signal.item(),
                'loss_rec_spectrogram': loss_fft.item(),
            }


            if self.overall_iteration % self.cfg.TRAIN.SAVE_REPORT_ITERATION == 0 or k==0:
                self.print_losses_to_file(losses)
                self.plot_to_file(x_orig, x_art, x_rec, y_art)

            if self.overall_iteration % self.cfg.TRAIN.SAVE_MODEL_ITERATION == 0 or k==0:
                self.save_model()

            self.current_iteration += 1
            self.overall_iteration += 1

        self.current_epoch += 1


    def print_losses_to_file(self, losses):
        path = self.path_report_losses
        lr = self.optimizer.param_groups[0]['lr']
        printStr = f'{self.current_iteration}, {lr}, ' + ', '.join([f'{i:.8f}' for k, i in losses.items()]) + '\n'
        print(self.overall_iteration, self.current_epoch, self.current_iteration, losses)
        with open(path, 'a') as f:
            f.write(printStr)

    # (x_orig, x_rec, y_art, att)
    def plot_to_file(self, x_orig, x_art, x_rec, y_art):
        x_orig = x_orig[-1].detach().cpu().squeeze().numpy()
        x_rec = x_rec[-1].detach().cpu().squeeze().numpy()
        x_art = x_art[-1].detach().cpu().squeeze().numpy()
        y_art = y_art[-1].detach().cpu().squeeze().numpy()
        img_path = os.path.join(self.path_report_images, f"{self.cfg.NAME}_epoch_{self.current_epoch:05d}_iteration_{self.current_iteration:05d}_step_{self.overall_iteration:05d}")

        plt.figure(figsize=(12, 4))
        ax0 = plt.subplot(3, 1, 1)
        plt.plot(x_orig)
        plt.plot(x_rec)
        plt.subplot(3, 1, 2, sharex=ax0)
        plt.plot(x_art)
        plt.subplot(3, 1, 3, sharex=ax0)
        plt.plot(y_art)
        plt.savefig(img_path + '.png')
        plt.savefig(img_path + '.svg')
        plt.close()
        #plt.show()

    def save_model(self):
        PATH = os.path.join(self.path_report_models, f"{self.cfg.NAME}_epoch_{self.current_epoch:05d}_iteration_{self.current_iteration:05d}_step_{self.overall_iteration:05d}_ArtifactEraser.pt")
        torch.save(self.model.state_dict(), PATH)

    def train(self):
        self.do_epoch()


class Trainer_:
    def __init__(self, config):
        self.cfg = config
        self.model_name = self.cfg.NAME
        exec(
        f"self.DatasetTrain={self.cfg.TRAIN.DATASET.split('.')[-1]}({self.cfg.TRAIN.SIGNAL_LENGTH}, {self.cfg.TRAIN.EEG_GENERATORS_DCGAN}, {self.cfg.TRAIN.STIM_ARTIFACT})"
        )
        self.DatasetTrain._len = self.cfg.TRAIN.BATCH_SIZE * self.cfg.TRAIN.N_ITERATIONS


        self.path_report = os.path.join(self.cfg.TRAIN.PATH_REPORT, self.cfg.NAME + "_" + datetime.now().strftime("%Y_%m_%d_%H_%M_%S"))
        self.path_report_models = os.path.join(self.path_report, 'Models')
        self.path_report_images = os.path.join(self.path_report, 'Images')
        self.path_report_config = os.path.join(self.path_report, 'config.yaml')
        self.path_report_losses = os.path.join(self.path_report, 'losses.csv')
        self.path_report_losses_validation = os.path.join(self.path_report, 'losses_val.csv')


        os.mkdir(self.path_report)
        os.mkdir(self.path_report_models)
        os.mkdir(self.path_report_images)
        with open(self.path_report_config, 'w') as file:
            yaml.dump(ObjDictToDict(self.cfg), file)


        with open(self.path_report_losses, 'w') as f:
            f.write('Epoch, lr, loss_rec_signal, loss_rec_spectrogram, loss_detection\n')

        self.save_model_freq = self.cfg.TRAIN.SAVE_MODEL_EPOCH
        self.save_report_freq = self.cfg.TRAIN.SAVE_REPORT_EPOCH

        self.GPUs = self.cfg.TRAIN.GPU
        self.num_gpus = self.GPUs.__len__()
        self.minibatchsize = int(self.cfg.TRAIN.BATCH_SIZE)

        self.lr = self.cfg.TRAIN.BASE_LR

        self.epochs = self.cfg.TRAIN.EPOCHS


        self.current_epoch = 0
        self.current_iteration = 0
        self.overall_iteration = 0
        self.epoch_list = []
        self.lossKL_list = []
        self.lossMSE_list = []
        self.loss_list = []
        self.device = self.GPUs[0]

        #self.model = GRU_Denoiser(n_filters=self.cfg.MODEL.ARCHITECTURE.N_FILTERS)

        exec(
        f"self.model={self.cfg.MODEL.MODEL.split('.')[-1]}(n_filters={self.cfg.MODEL.ARCHITECTURE.N_FILTERS})"
        )


        self.optimizer = optim.Adam(self.model.parameters(), lr=self.cfg.TRAIN.BASE_LR, betas=(self.cfg.TRAIN.BETA_1, 0.999))
        self.lr_scheduler = torch.optim.lr_scheduler.MultiStepLR(self.optimizer, milestones=self.cfg.TRAIN.DECAY_ITERATIONS, gamma=self.cfg.TRAIN.DECAY_FACTOR)

        self.criterionMSE = nn.MSELoss()
        self.criterionBCE = nn.BCELoss()

        def weights_init(m):
            classname = m.__class__.__name__
            if classname.find('Conv') != -1:
                nn.init.normal_(m.weight.data, 0.0, 0.02)
            elif classname.find('BatchNorm') != -1:
                nn.init.normal_(m.weight.data, 1.0, 0.02)
                nn.init.constant_(m.bias.data, 0)

        self.model.apply(weights_init)
        self.DLoaderTrain = DataLoader(self.DatasetTrain, batch_size=self.minibatchsize, shuffle=True, num_workers=self.cfg.TRAIN.CPU_COUNT_LOADERS, drop_last=False)
        self.model.cuda(self.device)


    def do_epoch(self):
        self.current_iteration = 0
        print('----------------------------------')
        print('VAE Epoch ' + str(self.current_epoch))


        self.model.train()
        #print('DOPICE')
        for k, (x_orig, x_art, y_art) in enumerate(self.DLoaderTrain):
            #break
            self.model.zero_grad()
            x_orig = x_orig.float().detach().to(self.device)
            x_art = x_art.float().detach().to(self.device)
            y_art = y_art.float().detach().to(self.device)

            x_rec, att = self.model(x_art)
            #x_orig_fft = torch.stft(x_orig.squeeze(1), n_fft=2*500, hop_length=100, return_complex=True, ).abs()
            #x_rec_fft = torch.stft(x_rec.squeeze(1), n_fft=2*500, hop_length=100, return_complex=True, ).abs()

            x_orig_fft = torch.fft.fft(x_orig.squeeze(1)).abs()
            x_rec_fft = torch.fft.fft(x_rec.squeeze(1)).abs()

            x_orig_fft = torch.log10(x_orig_fft+0.0001)
            x_rec_fft = torch.log10(x_rec_fft+0.0001)

            x_rec_fft[torch.isnan(x_orig_fft) | torch.isinf(x_orig_fft)] = 0
            x_orig_fft[torch.isnan(x_orig_fft) | torch.isinf(x_orig_fft)] = 0

            x_orig_fft[torch.isnan(x_rec_fft) | torch.isinf(x_rec_fft)] = 0
            x_rec_fft[torch.isnan(x_rec_fft) | torch.isinf(x_rec_fft)] = 0


            #att[att < 0] = att[att < 0]*0
            #att[att > 1] = att[att > 1]/att[att > 1]


            loss_signal = self.criterionMSE(x_rec[~torch.isnan(x_rec)], x_orig[~torch.isnan(x_rec)]) * self.cfg.TRAIN.WEIGHT_SIGNAL_RECONSTRUCTION
            loss_fft = self.criterionMSE(x_rec_fft[~torch.isnan(x_rec_fft)], x_orig_fft[~torch.isnan(x_rec_fft)])* self.cfg.TRAIN.WEIGHT_STFT_RECONSTRUCTION
            loss_att = self.criterionBCE(att[~torch.isnan(att)], y_art[~torch.isnan(att)]) * self.cfg.TRAIN.WEIGHT_DETECTION


            loss = loss_signal + loss_att + loss_fft
            loss.backward()

            torch.nn.utils.clip_grad_value_(self.model.parameters(), 0.1)
            self.optimizer.step()
            self.lr_scheduler.step()


            losses = {
                'loss_rec_signal': loss_signal.item(),
                'loss_rec_spectrogram': loss_fft.item(),
                'loss_detection': loss_att.item(),
            }


            if self.overall_iteration % self.cfg.TRAIN.SAVE_REPORT_ITERATION == 0 or k==0:
                self.print_losses_to_file(losses)
                self.plot_to_file(x_orig, x_art, x_rec, y_art, att)

            if self.overall_iteration % self.cfg.TRAIN.SAVE_MODEL_ITERATION == 0 or k==0:
                self.save_model()

            self.current_iteration += 1
            self.overall_iteration += 1

        self.current_epoch += 1


    def print_losses_to_file(self, losses):
        path = self.path_report_losses
        lr = self.optimizer.param_groups[0]['lr']
        printStr = f'{self.current_iteration}, {lr}, ' + ', '.join([f'{i:.8f}' for k, i in losses.items()]) + '\n'
        print(self.overall_iteration, self.current_epoch, self.current_iteration, losses)
        with open(path, 'a') as f:
            f.write(printStr)

    # (x_orig, x_rec, y_art, att)
    def plot_to_file(self, x_orig, x_art, x_rec, y_art, yy_art):
        x_orig = x_orig[-1].detach().cpu().squeeze().numpy()
        x_rec = x_rec[-1].detach().cpu().squeeze().numpy()
        x_art = x_art[-1].detach().cpu().squeeze().numpy()
        y_art = y_art[-1].detach().cpu().squeeze().numpy()
        yy_art = yy_art[-1].detach().cpu().squeeze().numpy()
        img_path = os.path.join(self.path_report_images, f"{self.cfg.NAME}_epoch_{self.current_epoch:05d}_iteration_{self.current_iteration:05d}_step_{self.overall_iteration:05d}")

        plt.figure(figsize=(12, 4))
        ax0 = plt.subplot(3, 1, 1)
        plt.plot(x_orig)
        plt.plot(x_rec)
        plt.subplot(3, 1, 2, sharex=ax0)
        plt.plot(x_art)
        plt.subplot(3, 1, 3, sharex=ax0)
        plt.plot(y_art)
        plt.plot(yy_art)
        # plt.savefig(img_path + '.png')
        plt.savefig(img_path + '.svg')
        plt.close()
        #plt.show()

    def save_model(self):
        PATH = os.path.join(self.path_report_models, f"{self.cfg.NAME}_epoch_{self.current_epoch:05d}_iteration_{self.current_iteration:05d}_step_{self.overall_iteration:05d}_ArtifactEraser.pt")
        torch.save(self.model.state_dict(), PATH)

    def train(self):
        self.do_epoch()






# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
import numpy as np
from scipy.stats._multivariate import multivariate_normal_gen, multivariate_normal_frozen

class multivariate_normal_(multivariate_normal_frozen):
    def __init__(self, X, allow_singular=False, seed=None):
        X = X.copy().T
        cov = np.cov(X.T)
        mu = X.mean(axis=0)
        super().__init__(mu, cov, allow_singular, seed)

    def pdf(self, X):
        X = X.copy().T
        return super().pdf(X)




# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


import os
import sys

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import time
import unittest
import matplotlib.pyplot as plt
import pandas as pd
import scipy.io as sio
import scipy.signal as signal
import scipy.stats as stats
import matplotlib.pyplot as plt
import numpy as np

from torch.utils.data import DataLoader
from scipy.io import loadmat


class SingleDataset_ThreeSeconds:
    def __init__(self, path, categories_keep='', categories_remove=''):
        self.path = path
        self.df = pd.read_csv(os.path.join(path, 'segments.csv'))

        self.categories_keep = categories_keep
        self.categories_remove = categories_remove


        if categories_keep:
            bool_vect = None
            for c in self.categories_keep:
                if isinstance(bool_vect, type(None)):
                    bool_vect = self.df.category_name == c
                else:
                    bool_vect = bool_vect | (self.df.category_name == c)
            self.df = self.df.loc[bool_vect].reset_index(drop=True)

        if categories_remove:
            bool_vect = None
            for c in self.categories_keep:
                if isinstance(bool_vect, type(None)):
                    bool_vect = self.df.category_name != c
                else:
                    bool_vect = (bool_vect) & (self.df.category_name != c)
            self.df = self.df.loc[bool_vect].reset_index(drop=True)


        #self.df = self.df.loc[self.df.category_name == 'pathology'].reset_index(drop=True)

        self.categories = self.df.category_name.unique()

        self.b, self.a = signal.butter(3, 200, analog=False, output='ba', fs=5000)
        self.b_low, self.a_low = signal.butter(3, 0.5, analog=False, output='ba', fs=500, btype='high')


    def __len__(self):
        return self.df.__len__()

    def __getitem__(self, item):
        x = loadmat(os.path.join(self.path, self.df.iloc[item]['segment_id']+'.mat'))['data'].squeeze()

        x[np.isnan(x)] = 0

        x = signal.filtfilt(self.b, self.a, x)
        x = x[::10]
        x = signal.filtfilt(self.b_low, self.a_low, x)

        ystr = self.df.iloc[item].category_name
        y = int(np.where(self.categories==self.df.iloc[item].category_name)[0])
        return x.copy(), y, ystr


class MergeDataset_ThreeSeconds:
    def __init__(self, path, categories_keep='', categories_remove=''):
        if type(path) != list:
            path = [path]

        self.categories_keep = categories_keep
        self.categories_remove = categories_remove

        self.Dats = [SingleDataset_ThreeSeconds(p, self.categories_keep, self.categories_remove) for p in path]
        self.lens = [d.__len__() for d in self.Dats]
        self.categories = np.unique(np.concatenate([d.categories for d in self.Dats]))

    def __len__(self):
        return sum(self.lens)


    def __getitem__(self, item):
        didx = 0 # dataset index

        if item < self.__len__():
            while True:
                if item >= self.lens[didx]:
                    item -= self.lens[didx]
                    didx += 1
                else:
                    x, _, ystr = self.Dats[didx][item]
                    y = int(np.where(self.categories==ystr)[0])
                    return x, y, ystr


class DatasetTimeDomain:
    def __init__(self, paths=[], categories_keep=[], test_ratio=0.2):
        import random
        random.seed(0)
        np.random.seed(0)
        torch.manual_seed(0)

        self.paths = paths
        self.categories = list(categories_keep)
        self.key_vals = {}
        for c, k in zip(self.categories, np.arange(categories_keep.__len__(), dtype=np.int)):
            self.key_vals[c] = k

        self.state = 'train'
        self.test_ratio = test_ratio

        self.datasets = {}
        self.idxes = {}
        cumulative_indexes = [0]
        for c in self.categories:
            self.datasets[c] = MergeDataset_ThreeSeconds(self.paths, categories_keep=[c])
            cumulative_indexes += [cumulative_indexes[-1] + self.datasets[c].__len__()]
            self.idxes[c] = np.random.permutation(np.arange(cumulative_indexes[-2], cumulative_indexes[-1]))

    def __len__(self):
        idx = 0
        for k, it in self.idxes.items():
            arr = self._get_idxes(it)
            idx += arr.__len__()
        return idx

    def train(self):
        self.state = 'train'
        return self

    def eval(self):
        self.state = 'eval'
        return self

    def _get_idxes(self, arr):
        idx_ = int(round(arr.shape[0] * self.test_ratio))
        if self.state == 'train':
            return arr[idx_:]
        else:
            return arr[:idx_]

    def __getitem__(self, item):
        idx = 0
        for k, it in self.idxes.items():
            arr = self._get_idxes(it)
            if item >= idx and item < idx+arr.shape[0]:
                x = self.datasets[k][item - idx][0]#[500:1000]
                return x, self.key_vals[k], k
            idx += arr.shape[0]







# Copyright 2020-present, Mayo Clinic Department of Neurology - Laboratory of Bioelectronics Neurophysiology and Engineering
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


"""
 Tools for digital signal processing: fft filtering, low-frequency filtering utilizing downsampling etc.
"""

import numpy as np
import scipy.signal as signal
import scipy.fft as fft
from tqdm import tqdm

def get_datarate(x):
    """
    Calculates datarate based on NaN values

    Parameters
    ----------
    x : numpy ndarray / list

    Returns
    -------
    numpy ndarray

    flaot values 0-1 with relative NaN count per signal.
    """
    if isinstance(x, np.ndarray):
        return 1 - (np.isnan(x).sum(axis=1)  / (x.shape[1]))
    else:
        return [1-(np.isnan(x_).sum()/x_.squeeze().shape[0]) for x_ in x]

def decimate(x, fs, fs_new, cutoff=None, datarate=False):
    """
    Downsample signal with anti-aliasing filter (Butterworth - 16th order).
    Works for signals with NaN values - replaced by zero.
    Can return also data-rate. Can take matrix where signals are stacked in 0th dim.

    Parameters
    ----------
    x : np ndarray
        shape[n_signal, n_sample], shape[n_sample] - can process multiple signals.
    fs : float
        Signals fs
    fs_new : float
        Downsample to fs_new
    cutoff : float
        Elective cutoff freq. of anti-alias. filter. By default 2/3 of 0.5*fs_new
    datarate : bool
        If return datarate of signals

    Returns
    -------
    numpy ndarray / tuple

    """

    x = x.copy()
    if datarate is True:
        datarate = get_datarate(x)

    if isinstance(cutoff, type(None)):
        cutoff = fs_new / 3 # two 3rds of half-sampling

    b_multiple_signals = True
    if x.ndim == 1:
        b_multiple_signals = False
        x = x.reshape(1, -1)

    for idx in range(x.shape[0]):
        x[idx, np.isnan(x[idx, :])] = np.nanmean(x[idx, :])

    b, a = signal.butter(16, cutoff/(0.5*fs), 'lp', analog=False)
    #b, a = signal.butter(3, cutoff/(0.5*fs), 'lp', analog=False)
    #a = [1]
    #b = signal.firwin(100, cutoff/(0.5*fs), pass_zero=True)
    #b /= b.sum()




    x = signal.filtfilt(b, a, x, axis=1)

    n_resampled = int(np.round((fs_new / fs) * x.shape[1]))
    x = signal.resample(x, n_resampled, axis=1)

    if b_multiple_signals is False:
        x = x.squeeze()

    if datarate:
        return x, datarate
    return x
    # PLOT AMPLITUDE CHAR
    #w, h = signal.freqs(b, a)
    #w = 0.5 * fs * w / w.max()
    #plt.semilogx(w, 20 * np.log10(abs(h) / (abs(h)).max()))
    #plt.title('Butterworth filter frequency response')
    #plt.xlabel('Frequency [radians / second]')
    #plt.ylabel('Amplitude [dB]')
    #plt.margins(0, 0.1)
    #plt.grid(which='both', axis='both')
    #plt.axvline(125, color='green') # cutoff frequency
    #plt.show()

def nandecimate(x, fs, fs_new, cutoff=None, datarate=False):
    """
    Downsample signal with anti-aliasing filter (Butterworth - 16th order).
    Works for signals with NaN values - replaced by zero.
    Can return also data-rate. Can take matrix where signals are stacked in 0th dim.

    Parameters
    ----------
    x : np ndarray
        shape[n_signal, n_sample], shape[n_sample] - can process multiple signals.
    fs : float
        Signals fs
    fs_new : float
        Downsample to fs_new
    cutoff : float
        Elective cutoff freq. of anti-alias. filter. By default 2/3 of 0.5*fs_new
    datarate : bool
        If return datarate of signals

    Returns
    -------
    numpy ndarray / tuple

    """

    x = x.copy()
    if datarate is True:
        datarate = get_datarate(x)

    if isinstance(cutoff, type(None)):
        cutoff = fs_new / 3 # two 3rds of half-sampling

    b_multiple_signals = True
    if x.ndim == 1:
        b_multiple_signals = False
        x = x.reshape(1, -1)


    nans = np.isnan(x)
    for idx in range(x.shape[0]):
        chmean = np.nanmean(x[idx, :])
        x[idx, nans[idx, :]] = np.nanmean(nans[idx, :])

   #for idx in range(x.shape[0]):
    #    idxes = np.where(nans[idx, :])[0]
    #    chmean = np.nanmean(x[idx, :])
    #    for i in tqdm(list(idxes)):
    #        loc_mean = np.nanmean(x[idx, i-100:i+100])
    #        if np.isnan(loc_mean): loc_mean = chmean
    #        x[idx, i] = loc_mean


    #b, a = signal.butter(16, cutoff/(0.5*fs), 'lp', analog=False)
    #b, a = signal.butter(3, cutoff/(0.5*fs), 'lp', analog=False)
    a = [1]
    b = signal.firwin(30, cutoff/(0.5*fs), pass_zero=True)
    b /= b.sum()

    x = signal.filtfilt(b, a, x, axis=1)

    n_resampled = int(np.round((fs_new / fs) * x.shape[1]))
    x = signal.resample(x, n_resampled, axis=1)
    nans = signal.resample(nans, n_resampled, axis=1)
    x[nans > 0.5] = np.nan

    if b_multiple_signals is False:
        x = x.squeeze()

    if datarate:
        return x, datarate
    return x
    # PLOT AMPLITUDE CHAR
    #w, h = signal.freqs(b, a)
    #w = 0.5 * fs * w / w.max()
    #plt.semilogx(w, 20 * np.log10(abs(h) / (abs(h)).max()))
    #plt.title('Butterworth filter frequency response')
    #plt.xlabel('Frequency [radians / second]')
    #plt.ylabel('Amplitude [dB]')
    #plt.margins(0, 0.1)
    #plt.grid(which='both', axis='both')
    #plt.axvline(125, color='green') # cutoff frequency
    #plt.show()

def unify_sampling_frequency(x : list, sampling_frequency: list, fs_new=None) -> tuple:
    """
    Takes list of signals and list of frequencies and downsamples to the same sampling frequency.
    If all frequencies are same and fs_new is not specified, no operation performed. If not all frequencies are the same
    and fs_new is not specified, downsamples all signals on the lowest fs present in the list. If fs_new is specified,
    signals will be processed and downsampled on that frequency. If all sampling frequencies == fs_new, nothing is performed.

    Parameters
    ----------
    x : list
        list of numpy signals for downsampling
    sampling_frequency : list
        for each signal
    fs_new : float
        new sampling frequency

    Returns
    -------
    tuple - (numpy ndarray, new_freq)
    """

    b_process = False

    if not isinstance(x, list):
        raise TypeError('First variable must be list of numpy arrays')

    if not isinstance(sampling_frequency, (list, np.ndarray)):
        raise TypeError('Second parameter must be list or array of floats/integers')
    sampling_frequency = np.array(sampling_frequency)

    if x.__len__() != sampling_frequency.__len__():
        raise AssertionError('Length of a signal list must be same as length of sampling_frequency list')

    fs_in_set = np.unique(sampling_frequency)
    if isinstance(fs_new, type(None)):
        if fs_in_set.__len__() > 1:
            b_process = True
            fs_new = fs_in_set.min()
    else:
        if (sampling_frequency != fs_new).sum() > 0:
            b_process = True
        else:
            fs_new = fs_in_set.min()

    if b_process is True:
        for idx in range(x.__len__()):
            fs = sampling_frequency[idx]
            sig = x[idx]
            sig = decimate(sig, fs, fs_new)
            x[idx] = sig
            sampling_frequency[idx] = fs_new

    return x, fs_new

def fft_filter(X:np.ndarray, fs:float, cutoff:float, type:str='lp'):
    """
    FFT filter

    Parameters
    ----------
    X : numpy.ndarray
    fs : float
    cutoff : float
    type : str
        'lp' or 'hp'

    Returns
    -------
    numpy.ndarray

    """

    Xs = fft.fft(X)
    freq = np.linspace(0, fs, Xs.shape[0])
    pos = np.where(freq > cutoff)[0][0]

    if type == 'lp':
        X_new = Xs
        X_new[pos:-pos] = 0
    elif type == 'hp':
        X_new = np.zeros_like(Xs)
        X_new[pos:-pos] = Xs[pos:-pos]
    X = np.real(fft.ifft(X_new))
    return X

def buffer(x:np.ndarray, fs:float=1, segm_size:float=None, overlap:float = 0, drop:bool=True):
    """
    Buffer signal into matrix

    Parameters
    ----------
    x : np.ndarray
        Signal to be
    fs : float
        Sampling frequency
    segm_size : float
        Segment size in seconds
    overlap : float
        Overlap size in seconds
    drop : bool
        Drop last segment if True, else append zeros
    Returns
    -------
    buffered_signal : np.ndarray
    """

    if not isinstance(x, np.ndarray):
        pass
    if x.ndim != 1:
        pass

    if isinstance(segm_size, type(None)):
        return x

    n_segm = int(round(fs * segm_size))
    #n_overlap = int(round(fs * overlap))
    n_shift = int(round(fs * (segm_size - overlap)))
    idx = 0

    buffered_signal = []
    while idx+n_segm-n_shift < x.shape[0]:
        app = x[idx:idx+n_segm]
        if app.__len__() < n_segm:
            if drop == False:
                app = np.append(app, np.zeros(n_segm - app.__len__()))
            else:
                idx = 2**30
                app = None

        if not isinstance(app, type(None)):
            buffered_signal.append(app)
        idx += n_shift
    return np.array(buffered_signal)

def PSD(x:np.ndarray, fs:float, nperseg=None, noverlap=0, nfft=None):
    """
    Estimates PSD of an input signal or signals using Welch's method.
    If nperseg is None, the spectrum is estimated from the whole signal in a single window.

    Parameters
    ----------
    x : np.ndarray
        A single signal with a shape (n_samples) or set of signals with a shape (n_signals, n_shapes)
    fs : float
        Sampling frequency
    nperseg : int
        Number of samples for a segment
    noverlap : int
        Number of overlap samples.

    Returns
    -------
    freq : np.ndarray
        Frequency axis for estimated PSD
    psd : np.ndarray
        Power spectral density estimate

    """
    axis = x.ndim-1
    if isinstance(nperseg, type(None)):
        nperseg = x.shape[axis]

    freq, psd = signal.welch(
        x,
        fs=fs,
        window='hann',
        nperseg=nperseg,
        noverlap=noverlap,
        nfft=nfft,
        detrend='constant',
        return_onesided=True,
        scaling='density',
        axis=axis,
        average='mean'
    )

    return freq, psd



    #N = xbuffered.shape[1]
    #psdx = fft.fft(xbuffered, axis=1)
    #psdx = psdx[:, 1:int(np.round(N / 2)) + 1]

    #psdx = (1 / (fs * N)) * np.abs(psdx) ** 2
    #psdx[np.isinf(psdx)] = np.nan
    return #psdx

class LowFrequencyFilter:
    """
        Parameters
        ----------
        fs : float
            sampling frequency
        cutoff : float
            frequency cutoff
        n_decimate : int
            how many times the signal will be downsampled before the low frequency filtering
        n_order : int
            n-th order filter used for filtration
        dec_cutoff : float
            relative frequency at which the signal will be filtered when downsampled
        filter_type : str
            'lp' or 'hp'
        ftype : str
            'fir' or 'iir'


        .. code-block:: python

            LFFilter = LowFrequencyFilter(fs=fs, cutoff=cutoff_low, n_decimate=2, n_order=101, dec_cutoff=0.3, filter_type='lp')
            X_inp = np.random.randn(1e4)
            X_outp = LFilter(X_inp)
    """

    __version__ = '0.0.2'

    def __init__(self, fs=None, cutoff=None, n_decimate=1, n_order=None, dec_cutoff=0.3, filter_type='lp', ftype='fir'):
        self.fs = fs
        self.cutoff = cutoff
        self.n_decimate = n_decimate
        self.dec_cutoff = dec_cutoff
        self.filter_type = filter_type

        self.n_order = n_order
        self.ftype = ftype

        self.n_append = None

        self.design_filters()

    def design_filters(self):
        if self.ftype == 'fir':
            if isinstance(self.n_order, type(None)): self.n_order = 101
            self.n_append = (2 * self.n_order) * (2**self.n_decimate)

            self.a_dec = [1]
            self.b_dec = signal.firwin(self.n_order, self.dec_cutoff, pass_zero=True)
            self.b_dec /= self.b_dec.sum()

            self.a_filt = [1]
            self.b_filt = signal.firwin(self.n_order, 2 * self.cutoff / (self.fs/2**self.n_decimate), pass_zero=True)
            self.b_filt /= self.b_filt.sum()

        elif self.ftype == 'iir':
            if isinstance(self.n_order, type(None)): self.n_order = 3
            self.n_append = (2 * self.n_order) * (2**self.n_decimate)

            self.b_dec, self.a_dec = signal.butter(self.n_order, self.dec_cutoff, btype='low')
            self.b_filt, self.a_filt = signal.butter(self.n_order,  2 * self.cutoff / (self.fs/2**self.n_decimate), btype='low')

        else: raise AssertionError(f'[INPUT ERROR]: ftype must be \'iir\' or \'fir\'')

    def decimate(self, X):
        X = signal.filtfilt(self.b_dec, self.a_dec, X)
        return X[::2]

    def upsample(self, X):
        X_up = np.zeros(X.shape[0] * 2)
        X_up[::2] = X
        X_up = signal.filtfilt(self.b_dec, self.a_dec, X_up) * 2
        return X_up

    def filter_signal(self, X):
        # append for filter
        X = np.concatenate((np.zeros(self.n_append), X, np.zeros(self.n_append)), axis=0)

        # append to divisible by 2
        C = int(2**np.ceil(np.log2(X.shape[0] + 2*self.n_append))) - X.shape[0]
        X = np.append(np.zeros(C), X)

        for k in range(self.n_decimate):
            X = self.decimate(X)

        X = signal.filtfilt(self.b_filt, self.a_filt, X)

        for k in range(self.n_decimate):
            X = self.upsample(X)
        #X = self.upsample(X)

        X = X[self.n_append + C : -self.n_append]
        return X


    def __call__(self, X):
        # append for filter
        X_orig = X.copy()
        X = self.filter_signal(X)
        if self.filter_type == 'lp': return X
        if self.filter_type == 'hp': return X_orig - X

def resample(x, fsamp_old, fsamp_new):
    '''
    :param x: input signal
    :param fsamp_old: original sampling frequency
    :param fsamp_new: new sampling frequency
    :return: interpolated signal at required sampling positions
    '''

    if fsamp_old == fsamp_new:
        return x

    nans = np.isnan(x)
    var = np.nanstd(x)
    mu = np.nanmean(x)
    x = (x - mu) / var
    x[np.isnan(x)] = 0

    xp = np.linspace(0, 1, len(x))
    xi = np.linspace(0, 1, int(np.round(len(x)*fsamp_new/fsamp_old)))

    # xi = np.arange(0, len(x), fsamp_old / fsamp_new) / fsamp_old  # nove navzorkovana casova osa v sekundach
    # xp = np.arange(0, len(x)) / fsamp_old  # puvodne navzorkovana casova osa v sekundach
    x = np.interp(xi, xp, x)
    nans = np.interp(xi, xp, nans)


    x[nans >= 0.5] = np.NaN
    x = (x * var) + mu
    return x

def detrend(y, x=None, y2=None):
    if isinstance(x, type(None)):
        x = np.linspace(0, 1, y.shape[0])

    a = (y[0] - y[-1]) / (x[0] - x[-1])
    b = y[0] - (x[0]*a)
    if isinstance(y2, type(None)):
        return y - (x*a+b)

    return y - (x * a + b), y2 - (x*a+b)

def find_peaks(y):
    position = []
    value = []
    for k in range(1, y.__len__() -1):
        if y[k-1] < y[k] and y[k+1] < y[k]:
            position += [k]
            value += [y[k]]
    return np.array(position), np.array(value)





# Copyright 2020-present, Mayo Clinic Department of Neurology - Bioelectronics Neurophysiology and Engineering Laboratory
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

from scipy.io import loadmat
from best import DELIMITER
from best.files import get_files


class _stimulation_artifacts(dict):
    _keys = dict([
        ( '.'.join(f.split(DELIMITER)[-1].split('.')[:-1]), f )
        for f in get_files(DELIMITER.join(__file__.split(DELIMITER)[:-1]), 'mat')
        if not '.-' in f.split(DELIMITER)[-1]
    ])

    def keys(self):
        return self._keys.keys()

    def get_artifact(self, item):
        f = self._keys[item]
        tmp = loadmat(f)
        return {'X': tmp['X'][0], 'fs': tmp['fs'][0, 0]}

    def __getitem__(self, item):
        return self.get_artifact(item)

    def __call__(self, item):
        return self[item]

    def __str__(self):
        return 'Stimulation Artifacts: ' + str(list(self._keys.keys()))

    def __repr__(self):
        return self.__str__()

stimulation_artifacts = _stimulation_artifacts()

"""

"""








# Copyright 2020-present, Mayo Clinic Department of Neurology - Bioelectronics Neurophysiology and Engineering Laboratory
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


class ObjDict(dict):
    """
    Dictionary which you can access a) as a dict; b) as a struct with attributes. Can use both foo adding and deleting
    attributes resp items. Inherits from dict
    """

    def __init__(self, VT_={}):
        super().__init__(VT_)
        for key in VT_.keys():
            self.__setattr__(key, VT_[key])

    def __setitem__(self, key, value):
        if key in self:
            del self[key]

        super().__setitem__(key, value)
        if not key in self.__dir__():
            self.__setattr__(key, value)

    def __setattr__(self, key, value):
        if key in self.__dir__():
            self.__delattr__(key)

        super().__setattr__(key, value)
        if not key in self:
            self.__setitem__(key, value)

    def __delitem__(self, key):
        value = super().pop(key)
        try:
            super().pop(value, None)
        except: pass
        if key in dir(self):
            self.__delattr__(key)

    def __delattr__(self, key):
        super().__delattr__(key)
        if key in self:
            self.__delitem__(key)

    def __repr__(self):
        return f"{type(self).__name__}({super().__repr__()})"

    def __missing__(self, key):
        self[key] = ObjDict()
        return self[key]

    def __getattr__(self, item):
        if not item in self.__dir__():
            self.__missing__(item)
        return super().__getattribute__(item)

class TwoWayDict(dict):
    def __init__(self, d=None):
        if not isinstance(d, type(None)):
            if isinstance(d, dict):
                for k, v in d.items():
                    self[k] = v

    def __setitem__(self, key, value):
        # Remove any previous connections with these values
        if key in self:
            del self[key]
        if value in self:
            del self[value]
        dict.__setitem__(self, key, value)
        dict.__setitem__(self, value, key)

    def __delitem__(self, key):
        dict.__delitem__(self, self[key])
        dict.__delitem__(self, key)

    def __len__(self):
        return dict.__len__(self) // 2


def DicToObj(d):
    """
    Converts dictionary into object where, keys - converts keys into attributes
    """
    top = type('new', (object,), d)
    seqs = tuple, list, set, frozenset
    for i, j in d.items():
        if isinstance(j, dict):
            setattr(top, i, DicToObj(j))
        elif isinstance(j, seqs):
            setattr(top, i,
                    type(j)(DicToObj(sj) if isinstance(sj, dict) else sj for sj in j))
        else:
            setattr(top, i, j)
    return top


def DictToObjDict(d):
    """
    Converts dictionary into object where, keys - converts keys into attributes
    """
    if isinstance(d, dict):
        #print(d.keys())
        top = ObjDict(d)

        #print(top.keys())
        for key in d.keys():
            if isinstance(d[key], dict):
                tmp = DictToObjDict(d[key])
                del top[key]
                top[key] = tmp

        return top
    else: return d


def ObjDictToDict(d):
    """
    Converts dictionary into object where, keys - converts keys into attributes
    """
    if isinstance(d, ObjDict):
        #print(d.keys())
        top = dict(d)

        #print(top.keys())
        for key in d.keys():
            if isinstance(d[key], ObjDict):
                tmp = ObjDictToDict(d[key])
                del top[key]
                top[key] = tmp

        return top
    else: return d




# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


import torch
import torch.nn as nn


from best.deep_learning import ResLayer1D


class Generator(nn.Module):
    def __init__(self, n_features=32, n_filters=64):
        super(Generator, self).__init__()
        self.n_features = n_features
        self.n_filters = n_filters
        self.dummy_param = nn.Parameter(torch.zeros(1))

        self.main = nn.Sequential(
            # input (batch_size, n_features, n_seconds)
            nn.ConvTranspose1d(in_channels=self.n_features, out_channels=self.n_filters, kernel_size=5, stride=5, padding=0, bias=False),
            nn.BatchNorm1d(self.n_filters),
            nn.ReLU(True),

            # input (batch_size, n_filters, n_seconds*5)
            nn.ConvTranspose1d(in_channels=self.n_filters, out_channels=self.n_filters, kernel_size=20, stride=10, padding=5, bias=False),
            nn.BatchNorm1d(self.n_filters),
            nn.ReLU(True),

            # input (batch_size, n_filters, n_seconds*50)
            nn.ConvTranspose1d(in_channels=self.n_filters, out_channels=self.n_filters, kernel_size=20, stride=10, padding=5, bias=False),
            nn.BatchNorm1d(self.n_filters),
            nn.ReLU(True),

            # input (batch_size, n_filters, n_seconds*500)
            ResLayer1D(self.n_filters, kernel_size=5, padding=2, bias=False),
            nn.BatchNorm1d(self.n_filters),
            nn.ReLU(True),

            # input (batch_size, n_filters, n_seconds*500)
            nn.ConvTranspose1d(in_channels=self.n_filters, out_channels=1, kernel_size=11, stride=1, padding=5, bias=False), # this was formerly 3
            # output (batch_size, 1, n_seconds*500)
            nn.ConvTranspose1d(in_channels=1, out_channels=1, kernel_size=21, stride=1, padding=10, bias=False),  # this was formerly 3
        )

    def forward(self, input):
        return self.main(input)

    def generate_signal(self, n_batch=1, n_seconds=1, momentum=0.5):
        noise = self._generate_noise(n_batch, n_seconds, momentum)
        x = self(noise).detach()
        x.detach().requires_grad = False
        return x

    def _generate_noise(self, n_batch=1, n_seconds=3, momentum=0.5):
        noise = [torch.randn(n_batch, self.n_features, 1, device=self.dummy_param.device)]
        # noise = [torch.randn(n_batch, 32, 1, device=self.dummy_param.device) * 0.10 - 5]
        for n in range(n_seconds - 1):
            noise += [
                momentum * noise[-1] + (1-momentum) * torch.randn(n_batch, self.n_features, 1, device=self.dummy_param.device)]
        noise = torch.cat(noise, dim=2)
        return noise


class Discriminator(nn.Module):
    def __init__(self, n_filters):
        super(Discriminator, self).__init__()
        self.dummy_param = nn.Parameter(torch.zeros(1))
        self.main = nn.Sequential(
            # input (batch_size, 1, n_seconds*500)
            nn.Conv1d(in_channels=1, out_channels=n_filters, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm1d(n_filters),
            nn.LeakyReLU(0.2, inplace=True),

            # input (batch_size, n_filters, n_seconds*500)
            ResLayer1D(n_filters, kernel_size=5, padding=2, bias=False),
            nn.BatchNorm1d(n_filters),
            nn.LeakyReLU(0.2, inplace=True),

            # input (batch_size, n_filters, n_seconds*500)
            nn.Conv1d(in_channels=n_filters, out_channels=n_filters, kernel_size=20, stride=10, padding=5, bias=False),
            nn.BatchNorm1d(n_filters),
            nn.LeakyReLU(0.2, inplace=True),

            # input (batch_size, n_filters, n_seconds*50)
            nn.Conv1d(in_channels=n_filters, out_channels=n_filters, kernel_size=20, stride=10, padding=5, bias=False),
            nn.BatchNorm1d(n_filters),
            nn.LeakyReLU(0.2, inplace=True),

            # input (batch_size, n_filters, n_seconds*5)
            nn.Conv1d(in_channels=n_filters, out_channels=1, kernel_size=5, stride=5, padding=0, bias=False),
            nn.Sigmoid()
            # output (batch_size, n_filters, n_seconds)
        )

    def forward(self, input):
        return self.main(input)


class DiscriminatorTimeFreq(nn.Module):
    def __init__(self, n_filters):
        super(DiscriminatorTimeFreq, self).__init__()
        self.dummy_param = nn.Parameter(torch.zeros(1))
        self.main_time = nn.Sequential(
            # input (batch_size, 1, n_seconds*500)
            nn.Conv1d(in_channels=1, out_channels=n_filters, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm1d(n_filters),
            nn.LeakyReLU(0.2, inplace=True),

            # input (batch_size, n_filters, n_seconds*500)
            ResLayer1D(n_filters, kernel_size=5, padding=2, bias=False),
            nn.BatchNorm1d(n_filters),
            nn.LeakyReLU(0.2, inplace=True),

            # input (batch_size, n_filters, n_seconds*500)
            nn.Conv1d(in_channels=n_filters, out_channels=n_filters, kernel_size=20, stride=10, padding=5, bias=False),
            nn.BatchNorm1d(n_filters),
            nn.LeakyReLU(0.2, inplace=True),

            # input (batch_size, n_filters, n_seconds*50)
            nn.Conv1d(in_channels=n_filters, out_channels=n_filters, kernel_size=20, stride=10, padding=5, bias=False),
            nn.BatchNorm1d(n_filters),
            nn.LeakyReLU(0.2, inplace=True),

            # input (batch_size, n_filters, n_seconds*5)
            nn.Conv1d(in_channels=n_filters, out_channels=1, kernel_size=15, stride=5, padding=0, bias=False),
            nn.Sigmoid()
            # output (batch_size, n_filters, n_seconds)
        )

        self.main_freq = nn.Sequential(
            # input (batch_size, 1, n_seconds*500)
            # nn.Conv1d(in_channels=1, out_channels=1, kernel_size=21, stride=1, padding=10, bias=False),
            nn.Conv1d(in_channels=1, out_channels=n_filters, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm1d(n_filters),
            nn.LeakyReLU(0.2, inplace=True),

            # input (batch_size, n_filters, n_seconds*500)
            ResLayer1D(n_filters, kernel_size=5, padding=2, bias=False),
            nn.BatchNorm1d(n_filters),
            nn.LeakyReLU(0.2, inplace=True),

            # input (batch_size, n_filters, n_seconds*500)
            nn.Conv1d(in_channels=n_filters, out_channels=n_filters, kernel_size=20, stride=10, padding=5, bias=False),
            nn.BatchNorm1d(n_filters),
            nn.LeakyReLU(0.2, inplace=True),

            # input (batch_size, n_filters, n_seconds*50)
            nn.Conv1d(in_channels=n_filters, out_channels=n_filters, kernel_size=20, stride=10, padding=5, bias=False),
            nn.BatchNorm1d(n_filters),
            nn.LeakyReLU(0.2, inplace=True),

            # input (batch_size, n_filters, n_seconds*5)
            nn.Conv1d(in_channels=n_filters, out_channels=1, kernel_size=15, stride=5, padding=0, bias=False),
            nn.Sigmoid()
            # output (batch_size, n_filters, n_seconds)
        )
        #
        # self.main_freq = nn.Sequential(
        #     nn.Linear(750, 100, bias=False),
        #     nn.BatchNorm1d(100),
        #     nn.LeakyReLU(0.2, inplace=True),
        #
        #     nn.Linear(100, 100, bias=False),
        #     nn.BatchNorm1d(100),
        #     nn.LeakyReLU(0.2, inplace=True),
        #
        #     nn.Linear(100, 1, bias=False),
        #     nn.BatchNorm1d(1),
        #     nn.Sigmoid()
        # )

        # self.main_merged = nn.Sequential(
        #     nn.Linear(200, 100, bias=False),
        #     nn.BatchNorm1d(100),
        #     nn.LeakyReLU(0.2, inplace=True),
        #
        #     nn.Linear(100, 1, bias=False),
        #     nn.BatchNorm1d(1),
        #     nn.Sigmoid()
        # )


    def forward(self, input):
        x_time = input[0]
        x_freq = input[1]
        return self.main_time(x_time).squeeze(), self.main_freq(x_freq).squeeze()
        # x_time_ = self.main_time(x_time).squeeze()
        # x_freq_ = self.main_freq(x_freq.squeeze())

        # x_ = torch.cat((x_time_, x_freq_), -1)
        # return self.main_merged(x_)






# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


from best import DELIMITER
from best._config import get_files
from torch import load, nn, zeros, unsqueeze, squeeze
import torch.nn.functional as F


TRAINED_MODELS = {'modelA': 'modelA_paper.pt', 'modelB': 'modelB_full.pt'}


def load_trained_model(model_name):
    if model_name in TRAINED_MODELS.keys():
        return _ModelsSeizureDetect.load_model(TRAINED_MODELS[model_name])
    else:
        raise KeyError(f"unknown trained model {model_name}; available {TRAINED_MODELS.keys()}")


class SeizureDetectModel(nn.Module):

    def __init__(self):
        super().__init__()
        # LSTM layer
        self.lstm_out = 100
        # kernel size
        self.nkernel = 20
        self.conv1 = nn.Conv2d(1, self.nkernel, (5, 5), padding=(0, 2))
        self.conv2 = nn.Conv2d(self.nkernel, self.lstm_out*2, (96, 3), padding=(0, 1))
        self.lstm = nn.LSTM(self.lstm_out*2, self.lstm_out, 2, bidirectional=True, dropout=0.5)
        self.fc1 = nn.Linear(2 * self.lstm_out, 4)

    def forward(self, x):
        bs = x.shape[0]
        h0 = zeros(2 * 2, bs, self.lstm_out).to(x.device)
        c0 = zeros(2 * 2, bs, self.lstm_out).to(x.device)
        x = unsqueeze(x, 1)
        x = F.relu(self.conv1(x))
        x = F.dropout(x, p=0.5, training=self.training)
        x = F.relu(self.conv2(x))
        x = F.dropout(x, p=0.5, training=self.training)
        x = squeeze(x)
        if bs == 1:
            x = unsqueeze(x, 0)
        x = x.permute(2, 0, 1)
        x, (hn, cn) = self.lstm(x, (h0, c0))
        x = F.dropout(x, p=0.5, training=self.training)
        x = self.fc1(x)
        return x, F.softmax(x, dim=2)


class _ModelsSeizureDetect(dict):
    _keys = dict([
        (
            '_'.join(f.split(DELIMITER)[-1].split('_')[:2]),
            f
        )
        for f in get_files(DELIMITER.join(__file__.split(DELIMITER)[:-1]), 'pt')
        if not '.-' in f.split(DELIMITER)[-1]
    ])

    def keys(self):
        return self._keys.keys()

    def get_model(self, item):
        f = self._keys[item]
        state_dict = load(f, map_location='cpu')
        mod = SeizureDetectModel()
        mod.load_state_dict(state_dict, strict=True)
        mod.eval()
        return mod

    @classmethod
    def load_model(cls, model_name):
        f = cls._keys[model_name]
        state_dict = load(f, map_location='cpu')
        mod = SeizureDetectModel()
        mod.load_state_dict(state_dict, strict=True)
        mod.eval()
        return mod

    def __getitem__(self, item):
        return self.get_model(item)

    def __call__(self, item):
        return self[item]

    def __str__(self):
        return 'models ' + str(self._keys.keys())

    def __repr__(self):
        return self.__str__()


# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


import torch
import torch.nn as nn
import torch.nn.functional as F


class ResLayer1D(nn.Module):
    def __init__(self, n=16, kernel_size=3, padding=1, bias=True):
        super(ResLayer1D, self).__init__()
        self.conv1 = nn.Conv1d(in_channels=n, out_channels=n, kernel_size=kernel_size, stride=1, padding=padding, bias=bias)
        self.conv2 = nn.Conv1d(in_channels=n, out_channels=n, kernel_size=kernel_size, stride=1, padding=padding, bias=bias)

    def forward(self, x):
        xe = self.conv1(x)
        xe = F.relu(xe, inplace=False)
        xe = self.conv2(xe)
        xe = F.relu(xe + x)
        return xe


class ResLayer2D(nn.Module):
    def __init__(self, n=16, kernel_size=3, padding=1, bias=True):
        super(ResLayer2D, self).__init__()
        self.conv1 = nn.Conv2d(in_channels=n, out_channels=n, kernel_size=kernel_size, stride=1, padding=padding, bias=bias)
        self.conv2 = nn.Conv2d(in_channels=n, out_channels=n, kernel_size=kernel_size, stride=1, padding=padding, bias=bias)

    def forward(self, x):
        xe = self.conv1(x)
        xe = F.relu(xe, inplace=False)
        xe = self.conv2(xe)
        xe = F.relu(xe + x, inplace=False)
        return xe


class Attention1D_MIL(nn.Module):
    def __init__(self, inp_size, downsample=2):
        super(Attention1D_MIL, self).__init__()

        self.conv_tanh = nn.Conv1d(in_channels=inp_size, out_channels=inp_size, kernel_size=1)
        self.conv_sigm = nn.Conv1d(in_channels=inp_size, out_channels=inp_size, kernel_size=1)
        self.conv_w = nn.Conv1d(in_channels=inp_size, out_channels=1, kernel_size=1)
        self.conv_final = nn.Conv1d(in_channels=inp_size, out_channels=1, kernel_size=1, stride=1)
        self.conv_donwsample = nn.Conv1d(in_channels=inp_size, out_channels=inp_size, kernel_size=downsample, stride=downsample, groups=inp_size, bias=False)

    def forward(self, x):
        tanh = torch.tanh(self.conv_tanh(x))
        sigm = torch.sigmoid(self.conv_sigm(x))
        z = self.conv_w(tanh * sigm)
        att = self.conv_final(x) * z
        x_weighted = att * x
        return self.conv_donwsample(x_weighted)#, att




# Copyright 2020-present, Mayo Clinic Department of Neurology - Laboratory of Bioelectronics Neurophysiology and Engineering
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import scipy.stats as stats

def tkeo(x):
    """
    Teager-Kaiser Energy Operator

    :param x: numpy ndarray
    :return: numpy ndarray
    """
    return (x[1:-1]**2) - (x[2:]*x[:-2])






# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


import numpy as np
from best.dbs import stimulation_artifacts

class ArtifactGenerator:
    def __init__(self, artifact_key='RCS', target_fs=500):
        tmp = stimulation_artifacts[artifact_key]
        self.artifact = tmp['X']
        self.fs_artifact = tmp['fs']
        self.fs_target = target_fs
        self.artifact_key = artifact_key

    def get_signal_random(self, n_batch, n_length, prob=None):
        sig_gen = []
        ref_gen = []

        for n in range(n_batch):
            if prob is None:
                _prob = np.random.rand()
            else:
                _prob = prob

            x = np.zeros(n_length)
            y = np.zeros(n_length)

            s = 0
            x_art = self.get_artifact()
            while s < x.shape[0] - 1 - x_art.shape[0] - x_art.shape[-1]-1:
                if np.random.rand() > (1 - _prob):
                    x_art = self.get_artifact()
                    s = s + np.random.randint(x_art.shape[0])
                    x[s:s + x_art.shape[0]] += x_art
                    y[s:s + x_art.shape[0]] = 1
                s = s + x_art.shape[0] + 3

            sig_gen += [x]
            ref_gen += [y]
        return np.stack(sig_gen).reshape(n_batch, 1, n_length), np.stack(ref_gen).reshape(n_batch, 1, n_length)

    def get_signal_periodic(self, n_batch, n_length, freq, ampl_rel):
        sig_gen = []
        ref_gen = []

        x_art = self._get_artifact()

        fs = self.fs_target
        stim_positions = np.arange(5, n_length-x_art.shape[0], fs/freq, dtype=np.int)

        for n in range(n_batch):
            x = np.zeros(n_length)
            y = np.zeros(n_length)


            for s in stim_positions:
                x_art = self._get_artifact() * ampl_rel
                x[s:s + x_art.shape[0]] += x_art
                y[s:s + x_art.shape[0]] = 1
            x /= self.artifact.max()

            sig_gen += [x]
            ref_gen += [y]
        return np.stack(sig_gen).reshape(n_batch, 1, n_length), np.stack(ref_gen).reshape(n_batch, 1, n_length)


    def get_artifact(self):
        temp = self._get_artifact()

        if np.random.randn() < 0:
            temp = temp * -1

        # f = np.random.rand() * 10 + 0.5
        f = 10 ** (np.random.rand() * (1.6 - (-1)) - 2)
        temp = temp * f
        if temp.shape[0] == 0: temp = np.append(temp[0])
        return temp

    def _get_artifact(self):
        xtemp = self.artifact
        len = int(np.round(self.fs_artifact / self.fs_target))

        s = np.random.randint(len)
        positions = np.arange(s, self.artifact.shape[0], len).astype(int)
        positions_m = (positions + np.random.randint(-len / 2, len / 2, positions.shape[0])).astype(int)
        positions_m[positions_m < 0] = 0
        positions_m[positions_m >= self.artifact.shape[0]] = self.artifact.shape[0] - 1

        temp = xtemp[positions_m]
        temp = temp - temp[0]
        return temp








# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
from scipy.signal import spectrogram
from scipy.stats import zscore
from torch import from_numpy

from best.deep_learning._models import load_trained_model

"""
    Seizure detection module 
    available trained models - 'modelA', 'modelB'
    modelA is the model from the published work.
    modelB had extended training dataset. 
    Optimal input for the model is 300 second. It is recommended to use only middle part of the signal.
    
Example
^^^^^^^^^

.. code-block:: python
    # load model
    modelA =  best.deep_learning.seizure_detect.load_trained_model('modelA')
    # load data
    fs = 500
    x_len = 300
    channels = 3
    # create fake data
    x_input = rand(channels, fs * x_len)
    # preprocess; from raw data to spectrogram
    x = best.deep_learning.seizure_detect.preprocess_input(x_input, fs)
    # get seizure probability; model has 4 output classes, seizure probability is class 4.
    # output is in shape (batch_size, x_len * 2 - 1); probability value for every half-second
    y = best.deep_learning.seizure_detect.infer_seizure_probability(x, modelA)

Sources
^^^^^^^^^^^

The seizure detection and training of the model is described in add website.
"""


def preprocess_input(x, fs):
    """
    This function will calculate a spectrogram. The spectrogram has shape [batch_size, 100, len(x) in seconds * 2 - 1]
    :param x: raw data input in bach form [batch_size, n-samples]
    :type x: iterable
    :param fs: sampling rate of the input signal
    :return: batch of spectrograms from the input
    :rtype: np.array
    """
    x = np.array(x)
    ii, jj = x.shape
    if np.sum(np.isnan(x[0, :])) < 0.75 * jj:
        mu = np.nanmean(x, axis=1)
        std = np.nanstd(x, axis=1)
        x = (x - mu.reshape((ii, 1))) / std.reshape((ii, 1))
        x = np.nan_to_num(x)
        f, t, x = spectrogram(x, fs, nperseg=fs, noverlap=fs / 2, axis=1)
        x2 = x[:, :100, :]
        x = np.empty(x2.shape)
        for kk, xx in enumerate(x2):
            idx = np.sum(xx, axis=0) == 0
            xx[:, ~idx] = zscore(xx[:, ~idx], axis=1)
            x[kk, :, :] = xx
        return x

    else:
        raise ValueError("too many NaN values in input")


def infer_seizure_probability(x, model, use_cuda=False, cuda_number=0):
    """
    infers seizure probability for a given input x; recommended signal len is 300 seconds.
    :param x: output from preprocess_input function, should be in shape [batch_size, 100, time_in_half-seconds]
    :param model: loaded seizure model
    :param use_cuda: if true x is loaded to cuda with cuda_number
    :param cuda_number:
    :return: seizure probability for input x in shape [batch_size, x.shape[2]]
    :rtype: np.array
    """
    x = from_numpy(x)
    batch_size = x.shape[0]
    if use_cuda:
        x = x.float().cuda(cuda_number)
    else:
        x = x.float()
    outputs, probs = model(x)
    probs = probs[:, :, 3].data.cpu().numpy().flatten()
    y = probs.reshape(x.shape[2], batch_size).T
    return y





import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from best.deep_learning.models_generic import ResLayer1D

from best.deep_learning import ResLayer1D

class dbs_artifact_removal_network_(nn.Module):
    def __init__(self, n_filters=64, fs=500):
        super().__init__()
        self.dummy_par = nn.Parameter(torch.zeros(1))

        self.conv1 = nn.Conv1d(in_channels=1, out_channels=16, kernel_size=3, stride=1, padding=1, bias=False)
        self.conv1_2 = nn.Conv1d(in_channels=16, out_channels=32, kernel_size=11, stride=1, padding=5, bias=False)
        self.gru_1 = nn.GRU(32, hidden_size=64, num_layers=1, bidirectional=False, bias=False, batch_first=True)
        self.conv1_3 = nn.Conv1d(in_channels=64, out_channels=1, kernel_size=3, stride=1, padding=1, bias=False)

        self.conv2 = nn.Conv1d(in_channels=1, out_channels=16, kernel_size=3, stride=1, padding=1, bias=False)
        self.conv2_1 = nn.Conv1d(in_channels=16, out_channels=32, kernel_size=11, stride=1, padding=5, bias=False)
        self.gru_encode = nn.GRU(32, hidden_size=64, num_layers=1, bidirectional=True, bias=False, batch_first=True)

        self.gru_decode = nn.GRU(32, hidden_size=64, num_layers=1, bidirectional=True, bias=False, batch_first=True)
        self.conv3 = nn.Conv1d(in_channels=128, out_channels=32, kernel_size=5, stride=1, padding=2, bias=False)
        self.resl3_1 = ResLayer1D(32, bias=False)
        self.convoutp = nn.Conv1d(in_channels=32, out_channels=1, kernel_size=11, stride=1, padding=5, bias=False)


    # def eval(self):
    #     super().eval()
    #     #self._opt = optim.Adam(self.parameters(), lr=0)
    #     self._loss = nn.MSELoss()
    #     return self

    def forward(self, x_inp):
        # x_inp = x_art

        x = x_inp
        x = self.conv1(x)
        x = F.relu(x)
        x = self.conv1_2(x)
        x = F.relu(x)
        x, tmp1 = self.gru_1(x.permute(0, 2, 1), torch.zeros(1, x_inp.shape[0], 64, dtype=torch.float, device=self.device))
        x = x.permute(0, 2, 1)
        x = self.conv1_3(x)
        # xatt =

        x_ = x_inp*(1-torch.relu(torch.sign(x)))
        # x_ = x_inp * (1 - torch.sigmoid(x))
        x_ = self.conv1(x_)
        x_ = F.relu(x_)
        x_ = self.conv1_2(x_)
        x_ = F.relu(x_)
        _, xemb = self.gru_encode(x_.permute(0, 2, 1), torch.zeros(2, x_inp.shape[0], 64, dtype=torch.float, device=self.device))
        x_dec, tmp2 = self.gru_decode(x_.permute(0, 2, 1), xemb)

        x_ = self.conv3(x_dec.permute(0,2,1))
        x_ = F.relu(x_)
        x_ = self.resl3_1(x_)
        x_outp = self.convoutp(x_)
        return x_outp, torch.sigmoid(x)# + x_outp

    @property
    def device(self):
        return self.dummy_par.device

class dbs_artifact_removal_network_v2(nn.Module):
    def __init__(self):
        super().__init__()
        self.dummy_par = nn.Parameter(torch.zeros(1))

        self.conv1_1 = nn.Conv1d(in_channels=3, out_channels=16, kernel_size=5, stride=1, padding=2, dilation=1, bias=False)
        self.conv1_2 = nn.Conv1d(in_channels=3, out_channels=16, kernel_size=5, stride=1, padding=6, dilation=3, bias=False)
        self.conv1_3 = nn.Conv1d(in_channels=3, out_channels=16, kernel_size=5, stride=1, padding=10, dilation=5, bias=False)
        self.conv1_4 = nn.Conv1d(in_channels=3, out_channels=16, kernel_size=5, stride=1, padding=20, dilation=10, bias=False)

        self.Res1 = ResLayer1D(64, 11, 5, bias=False)

        # self.gru_attention = nn.GRU(64, hidden_size=64, num_layers=1, bidirectional=False, bias=False, batch_first=True)
        self.conv2_1 = nn.Conv1d(in_channels=64, out_channels=64, kernel_size=3, stride=1, padding=1, bias=False)
        self.conv2_2 = nn.Conv1d(in_channels=64, out_channels=1, kernel_size=3, stride=1, padding=1, bias=False)

        # self.linear1 = nn.Linear(64, 64)
        # self.linear2 = nn.Linear(64, 64)
        # self.linear3 = nn.Linear(64, 64)

        self.lin1 = nn.Conv1d(in_channels=64, out_channels=64, kernel_size=3, stride=1, padding=1, bias=False)
        self.lin2 = nn.Conv1d(in_channels=64, out_channels=64, kernel_size=3, stride=1, padding=1, bias=False)
        self.lin3 = nn.Conv1d(in_channels=64, out_channels=64, kernel_size=3, stride=1, padding=1, bias=False)

        self.linear_logvar = nn.Linear(64, 1)
        self.linear_outp = nn.Linear(64, 1)

        # self.gru_decode = nn.GRU(64, hidden_size=64, num_layers=1, bidirectional=True, bias=False, batch_first=True)
        self.ResDecode = ResLayer1D(64, 11, 5, bias=False)

        self.conv3_1 = nn.Conv1d(in_channels=16, out_channels=1, kernel_size=5, stride=1, padding=2, dilation=1, bias=False)
        self.conv3_2 = nn.Conv1d(in_channels=16, out_channels=1, kernel_size=5, stride=1, padding=6, dilation=3, bias=False)
        self.conv3_3 = nn.Conv1d(in_channels=16, out_channels=1, kernel_size=5, stride=1, padding=10, dilation=5, bias=False)
        self.conv3_4 = nn.Conv1d(in_channels=16, out_channels=1, kernel_size=5, stride=1, padding=20, dilation=10, bias=False)


    def forward(self, x_inp):
        x = x_inp
        x = torch.cat(
            (
                self.conv1_1(x),
                self.conv1_2(x),
                self.conv1_3(x),
                self.conv1_4(x),
            ), 1
        )
        x = F.relu(x)
        x = self.Res1(x)

        # x_att, tmp1 = self.gru_attention(x.permute(0, 2, 1), torch.zeros(1, x_inp.shape[0], 64, dtype=torch.float, device=self.device))
        # x_att = x_att.permute(0, 2, 1)
        x_att = self.conv2_1(x)
        #x_att_outp = torch.sigmoid(self.conv2_2(x))

        logvar = self.linear_logvar(x_att.permute(0,2,1)).permute(0,2,1)
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(x)
        sample = (eps * std)

        x = \
            F.relu(self.lin1(x))\
            - F.relu(self.lin2((x_att * x))) \
            #+ F.relu(self.lin3((x_att * sample)))

        x_att_outp = torch.sigmoid(self.linear_outp(x_att.permute(0,2,1))).permute(0,2,1)


        # x = x.permute(0, 2, 1)

        # x_dec, tmp2 = self.gru_decode(x.permute(0, 2, 1))
        x_dec = x
        x_dec = self.ResDecode(x)

        x_dec = torch.cat(
            (
                self.conv3_1(x_dec[:, 0:16, :]),
                self.conv3_2(x_dec[:, 16:32, :]),
                self.conv3_3(x_dec[:, 32:48, :]),
                self.conv3_4(x_dec[:, 48:64, :]),
            ), 1
        ).sum(1).unsqueeze(1)

        return x_dec, x_att_outp

    @property
    def device(self):
        return self.dummy_par.device


class dbs_artifact_removal_network(nn.Module):
    def __init__(self, n_filters=64, fs=500):
        super().__init__()
        self.dummy_par = nn.Parameter(torch.zeros(1))

        if fs==500:
            self.conv1 = nn.Conv1d(in_channels=1, out_channels=n_filters, kernel_size=11, stride=2, padding=5, bias=False)
        elif fs==250:
            self.conv1 = nn.Conv1d(in_channels=1, out_channels=n_filters, kernel_size=5, stride=1, padding=2, bias=False)

        self.gru = nn.GRU(n_filters, hidden_size=128, num_layers=1, bidirectional=False, bias=False, batch_first=True)
        self.gru_attention = nn.GRU(n_filters, hidden_size=64, num_layers=1, bidirectional=False, bias=False, batch_first=True)

        self.conv2 = nn.Conv1d(in_channels=128+64, out_channels=n_filters, kernel_size=5, stride=1, padding=2, bias=False)
        self.resl2 = ResLayer1D(n_filters, bias=False)

        self.convoutp = nn.ConvTranspose1d(in_channels=n_filters, out_channels=1, kernel_size=12, stride=2, padding=5, bias=False)


        self.convfilter = nn.Conv1d(in_channels=1, out_channels=1, kernel_size=11, stride=1, padding=5, bias=False)

        #self.batch_outpatt = nn.BatchNorm1d(1)
        if fs == 500:
            self.convoutp_att = nn.ConvTranspose1d(in_channels=64, out_channels=1, kernel_size=12, stride=2, padding=5, bias=False)
        elif fs==250:
            self.convoutp_att = nn.ConvTranspose1d(in_channels=64, out_channels=1, kernel_size=6, stride=2, padding=2, bias=False)
    # def eval(self):
    #     super().eval()
    #     #self._opt = optim.Adam(self.parameters(), lr=0)
    #     self._loss = nn.MSELoss()
    #     return self


    def forward(self, x_inp):
        # x_inp = x_art

        x = x_inp
        x = self.conv1(x)
        x_features = F.relu(x)

        x, tmp1 = self.gru(x_features.permute(0, 2, 1), torch.zeros(1, x_inp.shape[0], 128, dtype=torch.float, device=self.device))
        x = x.permute(0, 2, 1)

        x_att, tmp2 = self.gru_attention(x_features.permute(0, 2, 1), torch.zeros(1, x_inp.shape[0], 64, dtype=torch.float, device=self.device))
        x_att = x_att.permute(0, 2, 1)

        x_ = torch.cat((x, x_att), dim=1)


        x_ = self.conv2(x_)
        x_ = F.relu(x_)
        x_ = self.resl2(x_)
        x_outp = self.convoutp(x_)
        x_outp = self.convfilter(x_outp)

        x_att = torch.sigmoid(self.convoutp_att(x_att))

        if not self.training:
            del x, x_features, tmp1, tmp2, x_


        return x_outp, x_att

    @property
    def device(self):
        return self.dummy_par.device



#
# class dbs_artifact_removal_network_(nn.Module):
#     def __init__(self, n_filters=64, fs=500):
#         super().__init__()
#         self.dummy_par = nn.Parameter(torch.zeros(1))
#
#         self.conv1 = nn.Conv1d(in_channels=1, out_channels=16, kernel_size=3, stride=1, padding=1, bias=False)
#         if fs==500:
#             self.conv1_2 = nn.Conv1d(in_channels=16, out_channels=32, kernel_size=11, stride=2, padding=5, bias=False)
#         elif fs==250:
#             self.conv1_2 = nn.Conv1d(in_channels=16, out_channels=32, kernel_size=5, stride=1, padding=2, bias=False)
#
#         # self.resl1_1 = ResLayer1D(32, kernel_size=21, padding=10, bias=False)
#         # self.conv1_3 = nn.Conv1d(in_channels=32, out_channels=32, kernel_size=11, stride=1, padding=5, bias=False)
#         self.resl1_2 = ResLayer1D(32, kernel_size=3, padding=1, bias=False)
#         # self.conv1_2 = nn.Conv1d(in_channels=32, out_channels=32, kernel_size=11, stride=1, padding=20, bias=False)
#         # self.resl1_3 = ResLayer1D(32, kernel_size=3, padding=1, bias=False)
#
#         self.gru_1 = nn.GRU(32, hidden_size=64, num_layers=1, bidirectional=False, bias=False, batch_first=True)
#         self.gru_2 = nn.GRU(64, hidden_size=64, num_layers=1, bidirectional=True, bias=False, batch_first=True)
#
#         self.gru_attention = nn.GRU(32, hidden_size=64, num_layers=1, bidirectional=False, bias=False, batch_first=True)
#
#         self.conv2 = nn.Conv1d(in_channels=128, out_channels=32, kernel_size=5, stride=1, padding=2, bias=False)
#         # self.resl2_1 = ResLayer1D(32, bias=False)
#         # self.resl2_2 = ResLayer1D(32, bias=False)
#         # self.resl2_3 = ResLayer1D(32, bias=False)
#
#         if fs == 500:
#             self.convoutp = nn.ConvTranspose1d(in_channels=32, out_channels=1, kernel_size=12, stride=2, padding=5, bias=False)
#         elif fs==250:
#             self.convoutp = nn.Conv1d(in_channels=32, out_channels=1, kernel_size=11, stride=1, padding=5, bias=False)
#
#
#         # self.conv3 = nn.Conv1d(in_channels=64, out_channels=32, kernel_size=5, stride=1, padding=2, bias=False)
#         # self.resl3_1 = ResLayer1D(64, bias=False)
#         # self.resl3_2 = ResLayer1D(32, bias=False)
#         # self.resl3_3 = ResLayer1D(32, bias=False)
#         # self.convoutp_att = nn.ConvTranspose1d(in_channels=32, out_channels=1, kernel_size=5, stride=1, padding=2, bias=False)
#
#         if fs == 500:
#             self.convoutp_att = nn.ConvTranspose1d(in_channels=64, out_channels=1, kernel_size=12, stride=2, padding=5, bias=False)
#         elif fs==250:
#             self.convoutp_att = nn.Conv1d(in_channels=64, out_channels=1, kernel_size=11, stride=1, padding=5, bias=False)
#
#         # self.convfilter = nn.Conv1d(in_channels=1, out_channels=1, kernel_size=11, stride=1, padding=5, bias=False)
#
#
#         #self.batch_outpatt = nn.BatchNorm1d(1)
#
#
#     # def eval(self):
#     #     super().eval()
#     #     #self._opt = optim.Adam(self.parameters(), lr=0)
#     #     self._loss = nn.MSELoss()
#     #     return self
#
#
#     def forward(self, x_inp):
#         # x_inp = x_art
#
#         x = x_inp
#         x = self.conv1(x)
#         x = F.relu(x)
#         x = self.conv1_2(x)
#         x = F.relu(x)
#
#         # x = self.resl1_1(x)
#         x_features = self.resl1_2(x)
#         # x_features = self.resl1_3(x)
#
#         x, tmp1 = self.gru_1(x_features.permute(0, 2, 1), torch.zeros(1, x_inp.shape[0], 64, dtype=torch.float, device=self.device))
#         x = x.permute(0, 2, 1)
#
#         x_att, tmp2 = self.gru_attention(x_features.permute(0, 2, 1), torch.zeros(1, x_inp.shape[0], 64, dtype=torch.float, device=self.device))
#         x_att = x_att.permute(0, 2, 1)
#         x_att = torch.sigmoid(x_att)
#
#         x_ = x - (x * x_att) + ((1-x_att) * torch.randn_like(x_att))
#
#         x_, tmp1 = self.gru_2(x_.permute(0, 2, 1), torch.zeros(2, x_inp.shape[0], 64, dtype=torch.float, device=self.device))
#         x_ = x_.permute(0, 2, 1)
#         #x_ = x - (x * x_att) + (x_att * torch.randn_like(x))
#         # x_ = x - x_att# + (torch.randn_like(x_att))
#
#         x_ = self.conv2(x_)
#         x_ = F.relu(x_)
#         # x_ = self.resl2_1(x_)
#         x_outp = self.convoutp(x_)
#
#         # x_att = self.conv3(x_att)
#         # x_att = F.relu(x_att)
#         # x_att = self.resl3_1(x_att)
#         x_att = self.convoutp_att(x_att)
#         x_att = torch.sigmoid(x_att)
#         # x_att = self.convoutp_att(x_att)
#
#         if not self.training:
#             del x, x_features, tmp1, tmp2, x_
#
#         return x_outp, x_att# + x_outp
#
#     @property
#     def device(self):
#         return self.dummy_par.device
#


#
# class dbs_artifact_removal_network_light(nn.Module):
#     def __init__(self, n_filters=64, fs=500):
#         super().__init__()
#         self.dummy_par = nn.Parameter(torch.zeros(1))
#
#         self.conv1 = nn.Conv1d(in_channels=1, out_channels=n_filters, kernel_size=11, stride=2, padding=5, bias=False)
#
#         self.gru = nn.GRU(n_filters, hidden_size=64, num_layers=1, bidirectional=False, bias=False, batch_first=True)
#
#         self.conv2 = nn.Conv1d(in_channels=64, out_channels=n_filters, kernel_size=5, stride=1, padding=2, bias=False)
#         self.convoutp = nn.ConvTranspose1d(in_channels=n_filters, out_channels=1, kernel_size=12, stride=2, padding=5, bias=False)
#         self.convfilter = nn.Conv1d(in_channels=1, out_channels=1, kernel_size=11, stride=1, padding=5, bias=False)
#
#
#     # def eval(self):
#     #     super().eval()
#     #     #self._opt = optim.Adam(self.parameters(), lr=0)
#     #     self._loss = nn.MSELoss()
#     #     return self
#
#     def forward(self, x_inp):
#         # x_inp = x_art
#
#         x = x_inp
#         x = self.conv1(x)
#         x_features = F.relu(x)
#
#         x, tmp1 = self.gru(x_features.permute(0, 2, 1), torch.zeros(1, x_inp.shape[0], 64, dtype=torch.float, device=self.device))
#         x = x.permute(0, 2, 1)
#
#         x_ = self.conv2(x)
#         x_ = F.leaky_relu(x_, 0.2)
#         x_outp = self.convoutp(x_)
#         x_outp = self.convfilter(x_outp)
#
#
#         if not self.training:
#             del x, x_features, tmp1, x_
#
#         return x_outp
#
#     @property
#     def device(self):
#         return self.dummy_par.device



# Copyright 2020-present, Mayo Clinic Department of Neurology - Laboratory of Bioelectronics Neurophysiology and Engineering
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


import numpy as np
from tqdm import tqdm

from best.files import create_folder, get_files, get_folders
from best.hypnogram.utils import time_to_timestamp
from best.cloud.mef import MefClient
from best import DELIMITER

from mef_tools.io import MefReader

from copy import deepcopy
from best.cloud.db import SystemStateLoader

from best.feature import augment_features, print_classification_scores, get_classification_scores
from best.feature_extraction.SpectralFeatures import mean_bands, mean_frequency, relative_bands
from best.dbs.artifact_removal import configs_ArtifactEraser, models_ArtifactEraser
from best.cloud.db import SessionFinder
from best.dbs.artifact_removal.trainer import *
from best.signal import get_datarate, buffer
from best.hypnogram.io import load_CyberPSG
import pandas as pd
from best.feature_extraction.FeatureExtractor import SleepSpectralFeatureExtractor
from sklearn.preprocessing import StandardScaler
from sklearn.svm import SVC
from sklearn.metrics import cohen_kappa_score, roc_curve, f1_score, accuracy_score
from best.modules import ZScoreModule, PCAModule
from best.hypnogram.utils import merge_annotations



class ReturnArrayObject:
    def __init__(self, parent, key):
        self.parent = parent
        self.key = key

    def __getitem__(self, item):
        # for d in self.parent.data:
        #     if d['channel_name'] == item:
        #         return d[self.key]

        if not isinstance(item, list):
            item = [item]

        dat = pd.DataFrame()
        for k, d in self.data.items():
            if d['channel_name'] in item:
                dat[k] = d['channel_name']
        return dat


    def __call__(self):
        return [d[self.key] for d in self.parent.data]


class EEGDataLoader:
    def __init__(self, pid, path_mef_files, path_annotations=None, db_id='UH3'):
        self.pid = pid
        self.db_id = db_id
        self.df_files = self.scan_mef_files(path_mef_files, pid)

        self.SSLoader = SystemStateLoader(self.db_id)

        if path_annotations:
            self.df_annotations = self.scan_hypnogram_annotations(path_annotations)
            self.df_annotations['path_mef'] = self.find_mef_sessions(self.df_annotations)

    def find_mef_sessions(self, df):
        print('Searching for sessions')
        # find a mef session for each annotation
        mef_files = []
        for idx, row in tqdm(list(df.iterrows())):
            s = row['start']
            e = row['end']

            file = self.df_files.loc[(s >= self.df_files.start) & (e <= self.df_files.end)]

            if file.__len__() == 0:
                pth = ''
            elif file.__len__() == 1:
                pth = file.iloc[0]['path']
            else: # if more sessions found, takes the one with more data
                pths = []
                drs = []

                for _, r in file.iterrows():
                    Rdr_ = MefReader(r['path'])
                    ch = [c for c in Rdr_.channels if not 'accl' in c][0]
                    x = Rdr_.get_data(ch, s * 1e6, e * 1e6)
                    pths += [r['path']]
                    drs += [1 - (np.isnan(x).sum() / x.shape[0])]
                    # ref_dr =
                pth = pths[np.argmax(drs)]
            mef_files += [pth]
        return mef_files

    def process_segment(self, s_process, e_process):
        process_segments = pd.DataFrame({'start': np.arange(s_process, e_process, 20 * 60)})
        process_segments['end'] = process_segments['start'] + 20 * 60
        files = self.find_mef_sessions(process_segments)
        process_segments['path_mef'] = files

        start = process_segments.start.min()
        end = process_segments.end.max()

        stim_info = self.SSLoader.load_stimulation_annotations(self.pid, start - 12 * 3600, end + 12 * 3600)
        if stim_info.__len__() == 0:
            stim_info = pd.DataFrame(
                [{'frequency': 0, 'amplitude': 0, 'start': start - 1 * 3600, 'end': end + 1 * 3600}])

        t = np.arange(stim_info.start.min(), stim_info.end.max(), 1)
        stim_ref = np.zeros_like(t) - 1

        for k, row_ in stim_info.iterrows():
            stim_ref[(t >= row_['start']) & (t < row_['end'])] = row_['frequency']



        process_segments_ = []
        for idx, row in process_segments.iterrows():

            # annot saying 0
            # -1 no annot
            # -2 annot saying NaN

            t_ = t[(t > row['start']) & (t < row['end'])]
            subsec = stim_ref[(t > row['start']) & (t < row['end'])]
            subsec[np.isnan(subsec)] = -2

            k_ = 0
            while k_ < subsec.__len__() - 1:
                start_freq = subsec[k_]
                start_time = t_[k_]
                while subsec[k_] == start_freq and k_ < subsec.__len__() - 1:
                    k_ += 1

                if start_freq == -1: start_freq = 0
                if start_freq == -2: start_freq = -1
                process_segments_ += [
                    {
                        'start': start_time,
                        'end': t_[k_],
                        'duration': t_[k_] - start_time,
                        'freq': start_freq,
                        'pid': self.pid,
                        'path_mef': row['path_mef']
                    }
                ]

        return pd.DataFrame(process_segments_)


    @staticmethod
    def scan_mef_files(path, pid=''):
        files_mef = get_files(path, 'mefd')

        df_files = []
        for pth in tqdm(files_mef):
            fid = pth.split(DELIMITER)[-1][:-5].split('_')
            do = True
            if pid:
                if fid[0] != pid:
                    do = False

            if do:
                Rdr = MefReader(pth)
                start = min(Rdr.get_property('start_time')) / 1e6
                end = max(Rdr.get_property('end_time')) / 1e6
                dur = (end - start)



                row = {
                    'pid': fid[0],
                    'fs': int(fid[2]),
                    'mode': fid[4],
                    'amplitude': float(fid[5][:-2]),
                    'start': start,
                    'end': end,
                    'duration': dur,
                    'artifacts_removed': bool(fid[7]),
                    'duration': dur,
                    'path': pth,
                    'channels': Rdr.channels
                }

                df_files += [row]
        df_files = pd.DataFrame(df_files)
        return df_files


    def scan_hypnogram_annotations(self, files_hyp):
        #files_hyp = [f for f in get_files(path, 'pseudogs_hypnogram.xml') if 'data_annotated' in f]

        df_annots = []
        for pth in tqdm(files_hyp):
            fid = pth.split(DELIMITER)[:-1]
            pid = fid[-4]

            files_mef = [f for f in get_files(DELIMITER.join(fid[:-1]), 'mefd') if
                         not 'features' in f.split(DELIMITER)[-1]]

            for fmef in files_mef:
                # print(fmef)
                Rdr = MefReader(fmef)
                fs = max(Rdr.get_property('fsamp'))
                channels = [ch for ch in Rdr.channels if not 'accel' in ch]

                # chans = np.unique(chans)

                mode = fid[8]

                if mode == 'no-stim':
                    mode = '0Hz'
                    amplitude = '0mA'

                elif 'sante' in mode:
                    amplitude = mode[-3:]
                    mode = 'sante145Hz'
                else:
                    amplitude = mode[-3:]
                    mode = mode[:-3]

                amplitude = float(amplitude[:-2])

                hyp = load_CyberPSG(pth)
                hyp = time_to_timestamp(hyp)
                hyp_start = hyp.start.min()
                hyp_end = hyp.end.max()

                row = {
                    'pid': fid[6],
                    'fs': fs,
                    'mode': mode,
                    'amplitude': amplitude,
                    'start': hyp_start,
                    'end': hyp_end,
                    'channels': channels,
                    'path': pth,
                }
                if not row in df_annots:
                    df_annots += [row]

        return pd.DataFrame(df_annots)


    def load_hypnogram_annotations(self):
        hyp_ = []
        for idx, row in self.df_annotations.iterrows():
            path_hyp = row['path']
            if path_hyp:
                hyp = load_CyberPSG(path_hyp)
                hyp = time_to_timestamp(hyp)
                hyp_start = hyp.start.min()
                hyp_end = hyp.end.max()

                stim_info = self.SSLoader.load_stimulation_annotations(self.pid, hyp_start - 12 * 3600, hyp_end + 12 * 3600)
                if stim_info.__len__() == 0:
                    stim_info = pd.DataFrame(
                        [{'frequency': 0, 'amplitude': 0, 'start': hyp_start - 1 * 3600, 'end': hyp_end + 1 * 3600}])

                t = np.arange(stim_info.start.min(), stim_info.end.max(), 1)
                stim_ref = np.zeros_like(t) - 1

                for k, row_ in stim_info.iterrows():
                    stim_ref[(t >= row_['start']) & (t < row_['end'])] = row_['frequency']

                # annot saying 0
                # -1 no annot
                # -2 annot saying NaN
                for k, row_hyp in hyp.iterrows():
                    t_ = t[(t > row_hyp['start']) & (t < row_hyp['end'])]
                    subsec = stim_ref[(t > row_hyp['start']) & (t < row_hyp['end'])]
                    subsec[np.isnan(subsec)] = -2

                    k_ = 0
                    while k_ < subsec.__len__() - 1:
                        start_freq = subsec[k_]
                        start_time = t_[k_]
                        while subsec[k_] == start_freq and k_ < subsec.__len__() - 1:
                            k_ += 1

                        if start_freq == -1: start_freq = 0
                        if start_freq == -2: start_freq = -1
                        hyp_ += [
                            {
                                'annotation': row_hyp['annotation'],
                                'start': start_time,
                                'end': t_[k_],
                                'duration': t_[k_] - start_time,
                                'pid': row['pid'],
                                'freq': start_freq,
                                'amplitude': row['amplitude'],
                                'pid': row['pid'],
                                'fs': row['fs'],
                                'channels': row['channels'],
                                'path_hypnogram': row['path'],
                                'path_mef': row['path_mef'],
                            }
                        ]

        hyp = pd.DataFrame(hyp_)
        return hyp









def compare_channels(x, channels):
    return sum([ch in channels for ch in x['channels']]) == channels.__len__()

def read_vaclav_table(path_csv):
    table = pd.read_csv(path_csv)
    table['DateTimeCompleted'] = table['DateTimeCompleted']

    time = []
    for v in table['DateTimeCompleted']:
        if not type(v) is str:
            v = None
        else:
            v = datetime.strptime(v, "%m/%d/%y %H:%M").timestamp()
        time += [v]
    table['DateTimeCompleted'] = time
    return table




class SignalSegment:
    def __init__(self, x=None, fs=None, channel_name=None, start=None, annotation=None):
        self.data = []
        if not isinstance(x, type(None)):
            self.add_channel(x, fs, channel_name, start, annotation)

        #self.fs = ReturnArrayObject(self, 'fs')
        #self.signal = ReturnArrayObject(self, 'data')
        #self.start = ReturnArrayObject(self, 'start')
        #self.end = ReturnArrayObject(self, 'end')
        #self.annotation = ReturnArrayObject(self, 'end')

    def add_channel(self, signal, fs, channel_name=None, start=None, annotation=''):
        if not channel_name:
            channel_name = 'channel_' + str(self.n_channels).zfill(3)

        if not start:
            start = 0

        end = start + (signal.shape[0] / fs)

        self.data += [{
            'channel_name': channel_name,
            'signal': signal,
            'fs': fs,
            'start': start,
            'end': end,
            'annotation': annotation
        }]


    @property
    def n_channels(self):
        return self.channels.__len__()

    @property
    def channels(self):
        return [d['channel_name'] for d in self.data]

    @property
    def annotation(self):
        kfunc = 'annotation'
        return self._get_attr(kfunc)

    @property
    def fs(self):
        kfunc = 'fs'
        return self._get_attr(kfunc)

    @property
    def start(self):
        kfunc = 'start'
        return self._get_attr(kfunc)

    @property
    def end(self):
        kfunc = 'end'
        return self._get_attr(kfunc)

    @property
    def signal(self):
        kfunc = 'signal'
        return self._get_attr(kfunc)


    def _get_attr(self, kfunc):
        dat = {}
        for d in self.data:
            if kfunc in d.keys():
                dat[d['channel_name']] = d[kfunc]
        return dat



class SignalRecording:
    def __init__(self):
        self.segments = []

    def add_segment(self, segment):
        self.segments += [segment]

    def __len__(self):
        return self.segments.__len__()

    def __getitem__(self, item):
        return self.segments[item]

    def extract_features(self, FtrExtractor):
        FtrRec = FeatureRecording()
        for idx in tqdm(range(self.__len__())):
            seg = self[idx]
            FtrSgm = FtrExtractor.extract_segment(seg)
            FtrRec.add_segment(FtrSgm)
        return FtrRec



class FeatureSegment:
    def __init__(self):
        self.data = {}
        self._datarate = {}
        self._feature_names = []

        self._must_be_keys = ['start', 'segment_duration', 'datarate']

    def add_channel(self, df_features, channel_name, feature_names):
        for k in self._must_be_keys:
            if not k in df_features.keys():
                raise KeyError(f"[KEY MISSING]: all this keys must be present in the df_features {self._must_be_keys}")

        self._feature_names = feature_names
        if not 'annotation' in df_features:
            df_features['annotation'] = ''
        self.data[channel_name] = df_features

    @property
    def datarate(self):
        dr = pd.DataFrame()
        for k, d in self.data.items():
            dr[k] = d['datarate']
        return dr

    @property
    def start(self):
        for k, d in self.data.items():
            return d['start']

    @property
    def segment_duration(self):
        for k, d in self.data.items():
            return d['segment_duration']

    @property
    def annotation(self):
        dr = pd.DataFrame()
        for k, d in self.data.items():
            dr[k] = d['annotation']
        return dr

    @property
    def n_channels(self):
        return self.data.keys().__len__()

    @property
    def channels(self):
        return list(self.data.keys())

    @property
    def feature_names(self):
        return self._feature_names

    def __getitem__(self, item):
        if isinstance(item, type(None)):
            item = self.channels
        if isinstance(item, str):
            item = [item]
        features = pd.DataFrame()
        features['start'] = self.start
        features['segment_duration'] = self.segment_duration
        features['datarate'] = self.datarate.min(axis=1)
        features['annotation'] = np.array([row.unique() for idx, row in self.annotation.iterrows()]).squeeze()

        feature_names = []
        for k, d in self.data.items():
            if k in item:
                for ftr_nm in self.feature_names:
                    fname = f"{k}-{ftr_nm}"
                    features[fname] = d[ftr_nm]
                    feature_names += [fname]
        return features, feature_names

    def select_channels(self, channels):
        new_self = deepcopy(self)
        new_self.data = dict([(ch, self.data[ch]) for ch in channels])
        return new_self


class FeatureRecording:
    def __init__(self):
        self.segments = []

    def add_segment(self, segment):
        self.segments += [segment]

    def __len__(self):
        return self.segments.__len__()

    def __getitem__(self, item):
        return self.segments[item]

    def __setitem__(self, key, value):
        self.segments[key] = value

    def __add__(self, other):
        self.segments += other.segments

    @property
    def channels(self):
        return list(np.unique( [self[k].channels for k in range(self.__len__()) ] ))

    def select_channels(self, channels):
        new_self = deepcopy(self)
        for k in range(self.__len__()):
            new_self[k] = new_self[k].select_channels(channels)
        return new_self


class PIBFeatureExtractor:
    def __init__(self,
                 fbands=[
                     [0.5, 5],  # delta
                     [4, 9],  # theta
                     [8, 14],  # alpha
                     [11, 16],  # spindle
                     [14, 20],
                     [20, 30]
                 ],
                 segment_size=30,
                 fs=250
                 ):

        self.fbands = fbands
        self.segment_size = segment_size
        self.fs = fs

        self.FeatureExtractor_MeanBand = SleepSpectralFeatureExtractor(
            fs=self.fs,
            segm_size=self.segment_size,
            fbands=self.fbands,
            datarate=False
        )

        self.FeatureExtractor_MeanBand._extraction_functions = \
            [
                mean_bands,
            ]

    def extract_segment(self, signal_segment):
        FtrSgm = FeatureSegment()
        for ch in signal_segment.channels:
            x = signal_segment.signal[ch]
            fs = signal_segment.fs[ch]
            s = signal_segment.start[ch]
            annotation = signal_segment.annotation[ch]

            if x.shape[0] / fs >= 30:
                x = buffer(x, fs=fs, segm_size=30, drop=True)
                datarate = get_datarate(x)
                stamps = np.arange(s, s + x.shape[0] * 30, 30)

                mean_bands, feature_names = self.FeatureExtractor_MeanBand(x)
                mean_bands = np.stack(mean_bands).T

                functions = [np.divide]
                symbols = ['/']
                mean_band_derived_features, mean_band_derived_names = mean_bands, feature_names
                for idx in range(functions.__len__()):
                    mean_band_derived_features, mean_band_derived_names = augment_features(
                        mean_band_derived_features,
                        feature_indexes=np.arange(mean_band_derived_features.shape[1]),
                        operation=functions[idx],
                        mutual=True, operation_str=symbols[idx],
                        feature_names=mean_band_derived_names

                    )


                mean_band_derived_features = np.log10(mean_band_derived_features)
                df = pd.DataFrame()
                df['start'] = stamps
                df['segment_duration'] = 30
                df['datarate'] = datarate
                df['annotation'] = annotation
                # print(mean_band_derived_features.shape, mean_band_derived_names.__len__(), df.shape, fs, self.FeatureExtractor_MeanBand._fs, x.shape)
                for f, fn in zip(mean_band_derived_features.T, mean_band_derived_names):
                    df[fn] = f

                FtrSgm.add_channel(
                    df_features=df,
                    channel_name=ch,
                    feature_names=mean_band_derived_names,
                )
        return FtrSgm


class MultiChannelSleepClassifier:
    __name__ = 'MultiChannelSleepClassifier'
    __version__ = '0.0.1'

    def __init__(self,
                 fbands=[
                     [0.5, 5],  # delta
                     [4, 9],  # theta
                     [8, 14],  # alpha
                     [11, 16],  # spindle
                     [14, 20],
                     [20, 30]
                 ],
                 segment_size=30,
                 fs=250,
                 dr_threshold = 0.8
                 ):

        self.fbands = fbands
        self.segment_size = segment_size
        self.fs = fs
        self.dr_threshold = dr_threshold

        self.FtrExtractor = PIBFeatureExtractor(
            fbands=fbands,
            segment_size=segment_size,
            fs=fs
        )

    def fit(self, X, y):
        self.scaler = StandardScaler()
        #self.UMAP = UMAP()
        self.PCA = PCAModule(var_threshold=0.95)
        Xe = self.scaler.fit_transform(X)
        Xe = self.PCA.fit_transform(Xe)

        #Xe = self.UMAP.fit_transform(Xe)
        self.clf = SVC(probability=True)
        self.clf.fit(Xe, y)

    def transform(self, X):
        x = self.scaler.transform(X)
        x = self.PCA.transform(x)
        # self.UMAP.transform(x)
        return x

    def scores(self, X):
        X = self.transform(X)
        return self.clf.predict_proba(X)

    def predict(self, X):
        return self.clf.predict(self.transform(X))

    def extract_features(self, SigRecording):
        return SigRecording.extract_features(self.FtrExtractor)

    def train(self, FtrRecording, test_ratio=0.2):
        channels = FtrRecording[0].channels
        self.channels = channels
        dr_threshold = self.dr_threshold

        df = []
        for k in range(FtrRecording.__len__()):
            df_, feature_names = FtrRecording[k][self.channels]
            df += [df_]

        df = pd.concat(df, ignore_index=True)
        df = df.loc[df.datarate >= dr_threshold]

        test_indexes = np.sort(np.concatenate(
            [df.loc[df.annotation == s].index[-int(np.floor(((df.annotation == s).sum() * test_ratio))):].to_numpy() for s in
             df.annotation.unique()]))
        train_indexes = [n for n in np.arange(df.__len__()) if not n in test_indexes and n in df.index]

        x_train = df.loc[train_indexes][feature_names].to_numpy()
        y_train = df.loc[train_indexes]['annotation'].to_numpy()

        x_test = df.loc[test_indexes][feature_names].to_numpy()
        y_test = df.loc[test_indexes]['annotation'].to_numpy()

        self.fit(x_train, y_train)
        yy_test = self.predict(x_test)

        results = {
            'kappa': {},
            'f1': {},
            'accuracy': {},
        }

        results['kappa']['all'] = cohen_kappa_score(y_test, yy_test)
        results['f1']['all'] = f1_score(y_test, yy_test, average='macro')
        results['accuracy']['all'] = accuracy_score(y_test, yy_test)

        for s in np.sort(np.unique(y_train)):
            results['kappa'][s] = cohen_kappa_score(y_test == s, yy_test == s)
            results['f1'][s] = f1_score(y_test == s, yy_test == s)
            results['accuracy'][s] = accuracy_score(y_test == s, yy_test == s)




        ret = {
            'train': {
                'x': x_train,
                'y': y_train
            },
            'test': {
                'x': x_test,
                'y': y_test,
                'yy': yy_test
            }
        }

        return results, ret


    def predict_FeatureRecording(self, FtrRecording):
        dr_threshold = self.dr_threshold

        df = []
        ftr_names = []
        for k in range(FtrRecording.__len__()):
            df_, feature_names = FtrRecording[k][self.channels]
            if feature_names.__len__() > 0: ftr_names = feature_names
            df += [df_]

        if df.__len__() > 0:
            df = pd.concat(df, ignore_index=True)
            df = df.loc[df.datarate >= dr_threshold]

            if df.__len__() == 0:
                return None

            bl = (df.sum(axis=1).isnull()) | (np.isinf(df.sum(axis=1).isnull())) | (np.isnan(df[feature_names]).sum(axis=1) > 0)
            df = df.loc[~bl].reset_index(drop=True)

            if df.__len__() == 0:
                return None

            x_predict = df[ftr_names].to_numpy()

            if x_predict.shape[0] == 0:
                return None

            yy_predict = self.predict(x_predict)

            hyp = df[['annotation', 'start', 'segment_duration']]
            hyp['annotation'] = yy_predict
            hyp['end'] = hyp['start'] + hyp['segment_duration']
            hyp = merge_annotations(hyp)
            return hyp





class CloudEEGDataLoader:
    def __init__(self, pid, db_id='UH3'):
        self.pid = pid
        self.db_id = db_id
        # modes - sante, continuous, nightoff

        self.SSLoader = SystemStateLoader(self.db_id)
        self.MfClient = MefClient()

        #if path_annotations:
            #self.df_annotations = self.scan_hypnogram_annotations(path_annotations)
            #self.df_annotations['path_mef'] = self.find_mef_sessions(self.df_annotations)

    def find_mef_sessions(self, df):
        print('Searching for sessions')
        # find a mef session for each annotation
        mef_files = []
        for idx, row in tqdm(list(df.iterrows())):
            s = row['start']
            e = row['end']

            file = self.df_files.loc[(s >= self.df_files.start) & (e <= self.df_files.end)]

            if file.__len__() == 0:
                pth = ''
            elif file.__len__() == 1:
                pth = file.iloc[0]['path']
            else: # if more sessions found, takes the one with more data
                pths = []
                drs = []

                for _, r in file.iterrows():
                    Rdr_ = MefReader(r['path'])
                    ch = [c for c in Rdr_.channels if not 'accl' in c][0]
                    x = Rdr_.get_data(ch, s * 1e6, e * 1e6)
                    pths += [r['path']]
                    drs += [1 - (np.isnan(x).sum() / x.shape[0])]
                    # ref_dr =
                pth = pths[np.argmax(drs)]
            mef_files += [pth]
        return mef_files

    def process_segment(self, s_process, e_process):

        SFndr = SessionFinder(self.db_id)

        process_segments = pd.DataFrame({'start': np.arange(s_process, e_process, 20 * 60)})
        process_segments['end'] = process_segments['start'] + 20 * 60
        process_segments['path_mef'] = SFndr.find_mef_session_bulk(self.pid, process_segments)
        process_segments = process_segments.loc[[p.__len__() > 0 for p in process_segments['path_mef'].to_list()]].reset_index(drop=True)
        process_segments = self.load_mef_info(process_segments)

        start = process_segments.start.min()
        end = process_segments.end.max()

        stim_info = self.SSLoader.load_stimulation_annotations(self.pid, start - 12 * 3600, end + 12 * 3600)
        if stim_info.__len__() == 0:
            stim_info = pd.DataFrame(
                [{'frequency': 0, 'amplitude': 0, 'start': start - 1 * 3600, 'end': end + 1 * 3600}])

        t = np.arange(stim_info.start.min(), stim_info.end.max(), 1)
        stim_ref = np.zeros_like(t) - 1
        ampl_ref = np.zeros_like(t) - 1
        for k, row_ in stim_info.iterrows():
            stim_ref[(t >= row_['start']) & (t < row_['end'])] = row_['frequency']
            ampl_ref[(t >= row_['start']) & (t < row_['end'])] = row_['amplitude']


        process_segments_ = []
        for idx, row in process_segments.iterrows():
            t_ = t[(t > row['start']) & (t < row['end'])]
            subsec = stim_ref[(t > row['start']) & (t < row['end'])]
            subsec[np.isnan(subsec)] = -2

            subsec_ampl = ampl_ref[(t > row['start']) & (t < row['end'])]
            subsec_ampl[np.isnan(subsec_ampl)] = 0

            k_ = 0
            while k_ < subsec.__len__() - 1:
                k0_ = k_
                start_freq = subsec[k_]
                start_time = t_[k_]
                while subsec[k_] == start_freq and k_ < subsec.__len__() - 1:
                    k_ += 1

                if start_freq == -1: start_freq = 0
                if start_freq == -2: start_freq = -1
                max_ampl = subsec_ampl[k0_:k_]
                max_ampl = np.max(max_ampl[~np.isnan(max_ampl)])
                
                process_segments_ += [
                    {
                        'start': start_time,
                        'end': t_[k_],
                        'duration': t_[k_] - start_time,
                        'freq': start_freq,
                        'ampl': max_ampl,
                        'pid': self.pid,
                        'path_mef': row['path_mef']
                    }
                ]
        return self.load_mef_info(pd.DataFrame(process_segments_))

    def load_mef_info(self, hyp):
        sessions_ = np.unique(hyp.path_mef.unique())
        sessions = pd.DataFrame()
        sessions['path'] = np.unique(sessions_)

        print('Searching for mef session details')
        fs = []
        channels = []
        for idx, row in tqdm(list(sessions.iterrows())):
            success, meta = self.MfClient.request_metadata(row['path'])
            fs_ = np.unique([ch_['fsamp'] for ch_ in meta if not 'acc' in ch_['name']])[0]
            channels_ = [ch_['name'] for ch_ in meta if not 'acc' in ch_['name']]

            fs += [fs_]
            channels += [channels_]

        sessions['channels'] = channels
        sessions['fs'] = fs

        fs = []
        channels = []
        for idx, row in hyp.iterrows():
            sr = sessions.loc[sessions.path == row['path_mef']]
            fs += [sr['fs']]
            channels += [sr['channels']]

        hyp['fs'] = fs
        hyp['channels'] = channels
        return hyp


    def load_stim_info(self, hyp):
        hyp_buffer = []

        hyp = time_to_timestamp(hyp)
        hyp_start = hyp.start.min()
        hyp_end = hyp.end.max()

        # Load stim info
        stim_info = self.SSLoader.load_stimulation_annotations(self.pid, hyp_start - 12 * 3600, hyp_end + 12 * 3600)
        if stim_info.__len__() == 0:
            hyp['freq'] = 0
            hyp['ampl'] = 0
            return hyp

        # prepare for tiling the hypnogram based on stim
        t = np.arange(stim_info.start.min(), stim_info.end.max(), 1)
        stim_ref = np.zeros_like(t) - 1
        ampl_ref = np.zeros_like(t) - 1
        for k, row_ in stim_info.iterrows():
            stim_ref[(t >= row_['start']) & (t <= row_['end'])] = row_['frequency']
            ampl_ref[(t >= row_['start']) & (t <= row_['end'])] = row_['amplitude']

        # tile hypnogram with stim annots
        for k, row_hyp in tqdm(list(hyp.iterrows())):
            t_ = t[(t >= row_hyp['start']) & (t <= row_hyp['end'])]

            subsec_freq = stim_ref[(t >= row_hyp['start']) & (t <= row_hyp['end'])]
            subsec_freq[np.isnan(subsec_freq)] = -2

            subsec_ampl = ampl_ref[(t >= row_hyp['start']) & (t <= row_hyp['end'])]
            subsec_ampl[np.isnan(subsec_ampl)] = 0
            # break

            k_ = 0
            done = False
            while k_ < subsec_freq.__len__() - 1:  # tiling each annotation based on the stim annotations
                done = True
                k0_ = k_
                start_time = t_[k_]
                start_freq = subsec_freq[k_]

                while subsec_freq[k_] == start_freq and k_ < subsec_freq.__len__() - 1:  # move till the stim freq is same
                    k_ += 1

                if start_freq == -1: start_freq = 0
                if start_freq == -2: start_freq = -1

                max_ampl = subsec_ampl[k0_:k_]
                max_ampl = np.max(max_ampl[~np.isnan(max_ampl)])

                r_ = row_hyp.to_dict()
                r_['start'] = start_time
                r_['end'] = t_[k_]
                r_['duration'] = t_[k_] - start_time
                r_['freq'] = start_freq
                r_['ampl'] = max_ampl

                hyp_buffer += [
                    r_
                ]

            if done == False:
                start_freq = 0
                max_ampl = 0
                r_ = row_hyp.to_dict()
                r_['freq'] = start_freq
                r_['ampl'] = max_ampl
                hyp_buffer += [
                    r_
                ]

        return pd.DataFrame(hyp_buffer)


    def load_hypnogram_annotations(self, files_hyp, round_timestamps=True):
        SFndr = SessionFinder(self.db_id)
        hyp = []
        for f in files_hyp:
            # Load hypnogram
            hyp_ = load_CyberPSG(f)
            hyp_ = time_to_timestamp(hyp_)
            if round_timestamps:
                hyp_['start'] = hyp_['start'].round()
                hyp_['end'] = hyp_['end'].round()
            hyp_['pid'] = self.pid
            hyp_['path_hypnogram'] = f
            hyp += [self.load_stim_info(hyp_)]

        hyp = pd.concat(hyp, ignore_index=True).sort_values('start').reset_index(drop=True)
        print('Searching for mef sessions')
        hyp['path_mef'] = SFndr.find_mef_session_bulk(self.pid, hyp)
        return hyp



def compare_channels(x, channels):
    return sum([ch in channels for ch in x['channels'].iloc[0]]) == channels.__len__()


def compare_freq(x, freqs):
    if freqs.__len__() == 0:
        return True
    return x['freq'] in freqs


def compare_ampl(x, freqs):
    if freqs.__len__() == 0:
        return True
    return np.ceil(x['ampl']) in freqs


def remove_artifacts(x, fs, cuda=3):
    model = models_ArtifactEraser[f'RCS_ArtifactEraser_{int(fs)}Hz_64']
    model = model.eval()
    model = model.cuda(cuda)

    mn_x = np.nanmean(x)
    x -= mn_x
    nans = np.isnan(x)

    fs = int(fs)
    shape = x.shape[0]
    x = np.concatenate((np.zeros(1 * fs), x, np.zeros(5 * fs)))

    xb = buffer(x, fs, segm_size=2, overlap=1.0)
    mn = np.nanmean(xb, axis=1)
    mn[np.isnan(mn)] = 0

    std = np.nanstd(xb, axis=1)
    std[np.isnan(std)] = 1

    dr = get_datarate(xb)
    xb[np.isnan(xb)] = 0

    xb -= mn.reshape(-1, 1)  # zscore
    xb /= std.reshape(-1, 1)

    dr_pos = np.where(dr > 0.1)[0]  # only datarate > than 0.1 in 30 minute segment
    xb_proc = xb[dr_pos]
    xb[:] = np.NaN

    # idx_end = 1
    # if xb_proc.shape[0]-512 > 1:
    #     idx_end = xb_proc.shape[0]-512
    idx_end = x.shape[0]


    with torch.no_grad():
        for k in np.arange(0, idx_end, 512):
            tmp = xb_proc[k:k + 512]
            bs = tmp.shape[0]
            if bs > 0:
                tmp_rec, det = model(torch.tensor(tmp).float().view(bs, 1, -1).cuda(cuda))
                tmp_rec = tmp_rec.detach().view(tmp_rec.shape[0], tmp_rec.shape[2]).cpu().numpy()
                xb_proc[k:k + bs] = tmp_rec

    xb[dr_pos] = xb_proc
    xb *= std.reshape(-1, 1)
    xb += mn.reshape(-1, 1)  # denormalize

    xrec = xb[:, int(0.5 * fs):int(1.5 * fs)].flatten()[
           int(0.5 * fs):int(0.5 * fs) + shape]  # crop only valid semgments
    xrec[nans] = np.nan
    x = xrec
    x += mn_x
    return x














# Copyright 2020-present, Mayo Clinic Department of Neurology - Laboratory of Bioelectronics Neurophysiology and Engineering
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import scipy.stats as stats


def mean_frequency(args):
    """
    **Mean dominant frequency**
    - Calculates mean dominant frequency on a frequency range defined as min-to-max of frequency bands at the input.
    - Source: https://www.mathworks.com/help/signal/ref/meanfreq.html

    Parameters
    ----------
    args : dictionary
        - 'psd' (*numpy.ndarray[n_samples, n_freq_samples]*) - one-sided PSD
        - 'fbands' (*list of lists*) - frequency bands in which the feature is to be calculated [[0.5, 4], [5, 9]]
        - 'freq' (*numpy.array[n_freq_samples]*) - reference frequency array for the PSD

    Returns
    -------
    x : list(numpy.array)
        Calculated features for individual frequency bands
    feature_name : list(numpy.array)
        Feature names

    """
    Pxx = args['psd']
    bands = args['fbands']
    freq = args['freq']

    f = args.freq

    min_position = np.nanargmin(np.abs(f - bands.min()))
    max_position = np.nanargmin(np.abs(f - bands.max()))

    P = Pxx[:, min_position: max_position + 1]
    f = f[min_position: max_position + 1]

    f = np.reshape(f, (1, -1))
    pwr = np.sum(P, axis=1)
    mnfreq = np.dot(P, f.T).squeeze() / pwr
    return [mnfreq], ['MEAN_DOMINANT_FREQUENCY']


def median_frequency(args):
    """
    **Spectral median frequency**

    - Calculates median dominant frequency on a frequency range defined as min-to-max of frequency bands at the input.
    - Source: https://www.mathworks.com/help/signal/ref/medfreq.html

    Parameters
    ----------
    args : dictionary
        - 'psd' (*numpy.ndarray[n_samples, n_freq_samples]*) - one-sided PSD
        - 'fbands' (*list of lists*) - frequency bands in which the feature is to be calculated [[0.5, 4], [5, 9]]
        - 'freq' (*numpy.array[n_freq_samples]*) - reference frequency array for the PSD

    Returns
    -------
    x : list(numpy.array)
        Calculated features for individual frequency bands
    feature_name : list(numpy.array)
        Feature names

    """
    Pxx = args['psd']
    bands = args['fbands']
    freq = args['freq']


    pwr = np.sum(Pxx, axis=1)
    #f = 0.5 * fs * np.arange(1, Pxx.shape[1]) / Pxx.shape[1]
    f = args.freq
    min_position = np.nanargmin(np.abs(f - bands.min()))
    max_position = np.nanargmin(np.abs(f - bands.max()))

    P = Pxx[:, min_position: max_position + 1]
    f = f[min_position: max_position + 1]

    pwr05 = np.repeat(pwr / 2, P.shape[1]).reshape(P.shape)
    P = np.cumsum(np.abs(P), axis=1)

    medfreq_pos = np.argmax(np.diff(P > pwr05, axis=1), axis=1) + 1
    medfreq = f.squeeze()[medfreq_pos]
    return [medfreq], ['SPECTRAL_MEDIAN_FREQUENCY']


def mean_bands(args):
    """
    **Mean power spectral density**
    - Mean spectral power for each frequency band.

    Parameters
    ----------
    args : dictionary
        - 'psd' (*numpy.ndarray[n_samples, n_freq_samples]*) - one-sided PSD
        - 'fbands' (*list of lists*) - frequency bands in which the feature is to be calculated [[0.5, 4], [5, 9]]
        - 'freq' (*numpy.array[n_freq_samples]*) - reference frequency array for the PSD

    Returns
    -------
    x : list(numpy.array)
        Calculated features for individual frequency bands
    feature_name : list(numpy.array)
        Feature names

    """
    Pxx = args['psd']
    bands = args['fbands']
    freq = args['freq']


    outp_params = []
    outp_msg = []
    for band in bands:
        subpsdx = Pxx[:, (freq >= band[0]) & (freq <= band[1])]
        outp_params.append(
            np.nanmean(subpsdx, axis=1)
        )
        outp_msg.append('MEAN_PSD' + str(band[0]) + '-' + str(band[1]) + 'Hz')
    return outp_params, outp_msg


def relative_bands(args):
    """
    **Relative spectral density**

    - Mean spectral power for each frequency band relative to the power of a whole spectrum defined on a frequency range defined as min-to-max of frequency bands at the input.

    Parameters
    ----------
    args : dictionary
        - 'psd' (*numpy.ndarray[n_samples, n_freq_samples]*) - one-sided PSD
        - 'fbands' (*list of lists*) - frequency bands in which the feature is to be calculated [[0.5, 4], [5, 9]]
        - 'freq' (*numpy.array[n_freq_samples]*) - reference frequency array for the PSD

    Returns
    -------
    x : list(numpy.array)
        Calculated features for individual frequency bands
    feature_name : list(numpy.array)
        Feature names

    """
    Pxx = args['psd']
    bands = args['fbands']
    freq = args['freq']


    outp_params = []
    outp_msg = []

    fullpsdx = np.nansum(Pxx[:, (freq >= bands.min()) & (freq <= bands.max())], axis=1)
    for band in bands:
        subpsdx = Pxx[:, (freq >= band[0]) & (freq <= band[1])]
        outp_params.append(
            np.nansum(subpsdx, axis=1) / fullpsdx
        )
        outp_msg.append('REL_PSD_' + str(band[0]) + '-' + str(band[1]) + 'Hz')
    return outp_params, outp_msg


def normalized_entropy(args):
    """
    **Spectral entropy (Shannon Entropy)**

    - Estimates Shannon Entropy of a spectrum on a frequency range defined as min-to-max of frequency bands at the input.
    - Source: https://www.mathworks.com/help/wavelet/ref/wentropy.html

    Parameters
    ----------
    args : dictionary
        - 'psd' (*numpy.ndarray[n_samples, n_freq_samples]*) - one-sided PSD
        - 'fbands' (*list of lists*) - frequency bands in which the feature is to be calculated [[0.5, 4], [5, 9]]
        - 'freq' (*numpy.array[n_freq_samples]*) - reference frequency array for the PSD

    Returns
    -------
    x : list(numpy.array)
        Calculated features for individual frequency bands
    feature_name : list(numpy.array)
        Feature names

    """
    Pxx = args['psd']
    bands = args['fbands']
    freq = args['freq']

    subpsdx = Pxx[:, (freq >= bands.min()) & (freq <= bands.max())]
    return [
               stats.entropy(subpsdx ** 2, axis=1)
           ], [
               'SPECTRAL_ENTROPY_' + str(bands.min()) + '-' + str(bands.max()) + 'Hz'
           ]


def non_normalized_entropy(args):
    """
    **Spectral entropy (Shannon Entropy)**

    - Estimates Shannon Entropy of a spectrum on a frequency range defined as min-to-max of frequency bands at the input.
    - Source: https://www.mathworks.com/help/wavelet/ref/wentropy.html

    Parameters
    ----------
    args : dictionary
        - 'psd' (*numpy.ndarray[n_samples, n_freq_samples]*) - one-sided PSD
        - 'fbands' (*list of lists*) - frequency bands in which the feature is to be calculated [[0.5, 4], [5, 9]]
        - 'freq' (*numpy.array[n_freq_samples]*) - reference frequency array for the PSD

    Returns
    -------
    x : list(numpy.array)
        Calculated features for individual frequency bands
    feature_name : list(numpy.array)
        Feature names

    """
    Pxx = args['psd']
    bands = args['fbands']
    freq = args['freq']

    subpsdx = Pxx[:, (freq >= bands.min()) & (freq <= bands.max())]
    return [
               - np.sum(subpsdx ** 2 * np.log(subpsdx ** 2), axis=1)
           ], [
               'SPECTRAL_ENTROPY_' + str(bands.min()) + '-' + str(bands.max()) + 'Hz'
           ]


def normalized_entropy_bands(args):
    """
    **Spectral entropy (Shannon Entropy)**

    - Estimates Shannon Entropy of a spectrum on a frequency range defined as min-to-max of frequency bands at the input.
    - Source: https://www.mathworks.com/help/wavelet/ref/wentropy.html

    Parameters
    ----------
    args : dictionary
        - 'psd' (*numpy.ndarray[n_samples, n_freq_samples]*) - one-sided PSD
        - 'fbands' (*list of lists*) - frequency bands in which the feature is to be calculated [[0.5, 4], [5, 9]]
        - 'freq' (*numpy.array[n_freq_samples]*) - reference frequency array for the PSD

    Returns
    -------
    x : list(numpy.array)
        Calculated features for individual frequency bands
    feature_name : list(numpy.array)
        Feature names

    """
    Pxx = args['psd']
    bands = args['fbands']
    freq = args['freq']


    outp_params = []
    outp_msg = []
    for band in bands:
        subpsdx = Pxx[:, (freq >= band[0]) & (freq <= band[1])]
        outp_params.append(
            stats.entropy(subpsdx ** 2, axis=1)
        )
        outp_msg.append('SPECTRAL_ENTROPY_' + str(band[0]) + '-' + str(band[1]) + 'Hz')
    return outp_params, outp_msg


def non_normalized_entropy_bands(args):
    """
    **Spectral entropy (Shannon Entropy)**

    - Estimates Shannon Entropy of a spectrum on a frequency range defined as min-to-max of frequency bands at the input.
    - Source: https://www.mathworks.com/help/wavelet/ref/wentropy.html

    Parameters
    ----------
    args : dictionary
        - 'psd' (*numpy.ndarray[n_samples, n_freq_samples]*) - one-sided PSD
        - 'fbands' (*list of lists*) - frequency bands in which the feature is to be calculated [[0.5, 4], [5, 9]]
        - 'freq' (*numpy.array[n_freq_samples]*) - reference frequency array for the PSD

    Returns
    -------
    x : list(numpy.array)
        Calculated features for individual frequency bands
    feature_name : list(numpy.array)
        Feature names

    """
    Pxx = args['psd']
    bands = args['fbands']
    freq = args['freq']


    outp_params = []
    outp_msg = []
    for band in bands:
        subpsdx = Pxx[:, (freq >= band[0]) & (freq <= band[1])]
        outp_params.append(
            - np.sum(subpsdx ** 2 * np.log(subpsdx ** 2), axis=1)
        )
        outp_msg.append('SPECTRAL_ENTROPY_' + str(band[0]) + '-' + str(band[1]) + 'Hz')
    return outp_params, outp_msg






# Copyright 2020-present, Mayo Clinic Department of Neurology
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.


from best import DELIMITER
from best._config import get_files, config

configs_DCGAN = dict([(f.split(DELIMITER)[-1].split('.')[0], config(f)) for f in get_files(DELIMITER.join(__file__.split(DELIMITER)[:-1]), 'yaml') if not '.-' in f.split(DELIMITER)[-1]])




